<!DOCTYPE html>
<html>
    <head>
        <meta http-equiv="Content-Type" content="text/html; charset=UTF-8">
        <meta http-equiv="Cache-Control" content="no-cache, no-store, must-revalidate" />
        <meta http-equiv="Pragma" content="no-cache" />
        <meta http-equiv="Expires" content="0" />
        <title>QuEST Local Control Panel</title>
        <link id="style_sheet" rel="stylesheet" href="Style/control-panel.css" type="text/css">
        <link rel="stylesheet" href="Style/jquery-ui-1.10.3.custom.css" type="text/css">
        <script type="text/javascript" src="js/jquery-1.9.1.js"></script>
        <script type="text/javascript" src="js/jquery-ui-1.10.3.custom.min.js"></script>
        <script type="text/javascript" src="js/kinetic-v5.1.0.min.js"></script>

        <script>

            //****global variables***************************************************************************************************************************************************************
            var QnetClientThread;                       //web worker object for qnet_client.js web worker thread that sends controls to and requests indications from the server.
            var linkOk = false;                         //link health between this client and the server... this flag is needed to detect transitions in link state
                                                        //that is determined by the web worker (qnet_client.js).
            var redrawComponents = false;               //flag set when component layer needs to be drawn/updated which will occur at a browser determined interval in requestAnimationFrame processing.
            var developerMode = false;                  //true when connected to the QLCP-NET Developer Tool server via localhost
            var designMode = false;                     //true when in design mode (able to edit panel layout and bit assignments)
            var componentStyle = true;                  //true when current panel design style is 'Component' and not 'Mosaic Tile'        
            var controlOptionsStr = "";                 //string of the control label/bit options used to set the innerHTML of properties select lists      
            var indicationOptionsStr = "";              //string of the indication label/bit options used to set the innerHTML of properties select lists   
            var activeIcons = new Array();              //this array stores the IDs of icons that are currently active (in the grid) for mosaic tile design style
            var selectedIcons = new Array();            //array of icon IDs for the currently selected icons (mosaic tile design style)
            var nbrControls = 10;                       //the number of controls that can be assigned to an icon (mosaic tile design style)
            var nbrRows = 60;                           //rows in editor grid
            var nbrCols = 120;                          //columns in editor grid
            var controlBits = new Array(512);           //this array stores control bit states for latched controls (e.g. toggle switch, keyswitch, icon controls when checkbox is checked)
            var requestFileName = false;                //the name of the panel file to open; it is requested from the developer tool (server)
            var hostUrl = window.location.host;         //IP address of the QLCP-NET server that is entered for the URL in the browser       
            var passcode = "";                          //passcode entered by user for locked controls (component design style)               
            var lastClickTime = new Date();             //this is the last time the user clicked on a control component (used for passcode functionality)
            var idNbr = 0;                              //the next id number to assign to an icon or component... this ensures all id numbers are unique
            var initLatchedSwitches = false;            //flag is true when changing between run and design mode to initialize latched switches
            var ctrlKeyDown = false;                    //capture when control key is pressed
            var shiftKeyDown = false;                   //capture when shift key is pressed                
            var backgroundColor = '#333333';            //background color of grid area (main content)
            var defaultTrackColor = '#eeeeee';          //default track color
            var defaultFillColor = '#555555';           //default fill color           
            var selectedId = "";                        //id of currently selected object (id is built from idNbr and prefix)
            var selectedComponents = new Array();       //array containing the selected component objects
            var compInSelectedGroup = false;            //true when component is in a selected group 
            var mouseDownOnComponent = false;           //true if there was a mouse down event on any component
            var initX = 0;                              //initial x coordinate of mouse position
            var initY = 0;                              //initial y coordinate of mouse position
            var deltaX = 0;                             //delta from initial x coordinate when moving a group of selected components
            var deltaY = 0;                             //delta from initial y coordinate when moving a group of selected components
            var draggingToSelectComponents = false;     //true while dragging mouse to select a group of components
            var selectionBoxCreated = false;            //set true when the rectangle shape for the selection box is added to the component layer
            var selectionBox = null;                    //selection box (rectangle shape) object displayed while making a group selection 
            
            //TODO: remove this variable and associated code since it is not currently used.
            //This was the attempt to support multi-icon selections. Going forward, use implementation for multiple component selection
            var selectionStart;                         //grid id of the start of a selection (mouse down)

            //This is an array of colors used for drawing icons, components, and text
            var iconColors = ["red", "yellow", "lime", "green", "orange", "blue", "white", "black", "cyan", "dodgerblue", "gray", "dimgray", "transparent"];
                
            //The div id number in the icon dialog box is used as an index into this array of icon types.
            var iconTypes = ["signal", "signal_flip_v", "signal_flip_h", "signal_180",
                "signal_90_flip_h", "signal_90", "signal_270", "signal_270_flip_h",
                "switch", "switch_flip_v", "switch_flip_h", "switch_180",
                "switch_90_flip_h", "switch_90", "switch_270", "switch_270_flip_h",
                "track", "track_90", "track_45", "track_135",
                "track_corner", "track_corner_flip_v", "track_corner_flip_h", "track_corner_180", 
                "track_corner_90_flip_h", "track_corner_90", "track_corner_270", "track_corner_270_flip_h",
                "track_cross", "track_cross_diag", "circle", "arrow", "arrow_90", "arrow_180", "arrow_270", "blank",
                "signal_45", "signal_135", "signal_225", "signal_315",
                "signal_45_flip_h", "signal_135_flip_h", "signal_225_flip_h", "signal_315_flip_h"];
                              
            //This is a custom data string that will allow an icon's data to be stored as a data attribute in its canvas element.
            //Data for controls will be stored as "button_label,index"
            //Data for indications will be stored as "index,state", where state is not the current state of the status but defines for which state
            //(true or false) the color should be displayed.
            var customDataStr = 'data-newdrop="1" data-gridlocation="" '
                + 'data-control0=",0" data-control1=",0" data-control2=",0" data-control3=",0" data-control4=",0" '
                + 'data-control5=",0" data-control6=",0" data-control7=",0" data-control8=",0" data-control9=",0" '
                + 'data-indication0="0" data-indication1="0" '
                + 'data-indication2="0" data-indication3="0" data-indication4="0"  data-indication5="0"  data-indication6="0"  data-indication7="0" '
                + 'data-indication8="0" data-indication9="0" data-indication10="0"  data-indication11="0"  data-indication12="0"  data-indication13="0" '
                + 'data-indication14="0" data-indication15="0" data-indication16="0"  data-indication17="0" '
                + 'data-indication18="0" data-indication19="0" data-indication20="0"  data-indication21="0" '
                + 'data-indication22="0" data-indication23="0" data-indication24="0"  data-indication25="0" '
                + 'data-indication26="0" data-indication27="0" '
                + 'data-defaultcolor1="' + defaultTrackColor + '" data-defaultcolor2="' + defaultTrackColor + '"';
        
            //Create key/value pairs of drop modes for each component type
            //The associated value is a boolean that is true when the drop mode for that component type is active.
            var dropMode = {
                text: false,
                LED: false,
                pb: false,
                pb_square: false,
                togSw: false,
                togSwMom: false,
                quadrilateral: false,
                arrow: false,
                timer: false,
                icons: false,
                keyswitch: false,
                none: true
            };
            
            //create key/value pairs that map button id to component type
            var buttonToComponent = {
                text_button: "text",
                led_button: "LED",
                pb_button: "pb",
                pb_square_button: "pb_square",
                togSw_button: "togSw",
                quadrilateral_button: "quadrilateral",
                arrow_button: "arrow",
                timer_button: "timer",
                icons_button: "icons",
                keyswitch_button: "keyswitch"
            };
            
            //array used for display date and time in the LCP header div
            var monthNames = ["JAN", "FEB", "MAR", "APR", "MAY", "JUN", "JUL", "AUG", "SEP", "OCT", "NOV", "DEC"];
            
            //***********************************************************************************************************************************************************************           
                             
            //after the document page has loaded and is ready...
            $(document).ready(function()  
            {   
                
                InitGrid();

                InitDialogBoxes();

                //Check if the browser supports web workers,
                //then create a web worker to handle the QLCP-NET client communications.
                if (typeof(Worker) !== "undefined")
                {                   
                    //Create the thread for QLCP-NET client communications.
                    QnetClientThread = new Worker("qnet_client.js");
                    
                    //Add the event listener to process indication messages received from the QLCP-NET client comms thread.
                    //Message data is received as msg.data object
                    QnetClientThread.addEventListener("message", function(msg){
                    
                        //if the link is down and then recovers...
                        if (!linkOk && msg.data.link)
                        {
                            //update the link status image
                            document.getElementById("link_status").src = "Images/linkOK24px.png";
                            linkOk = true;
                            SendAllControls();
                        }
                        
                        //if the link is healthy and then fails...
                        if (linkOk && !msg.data.link)
                        {
                            //update the link status image
                            document.getElementById("link_status").src = "Images/error24px.png";
                            linkOk = false;
                        }
                                             
                        //Since this event will be fired periodically (sub-second interval),
                        //it is a convenient time to update the system time displayed in the header.
                        var d = new Date();
                        var dateTimeStr = d.getFullYear()
                            + '-' + monthNames[d.getMonth()]
                            + '-' + pad(d.getDate())
                            + '  ' + pad(d.getHours())
                            + ':' + pad(d.getMinutes());

                        document.getElementById("system_time").innerHTML = dateTimeStr;   
                        
                        //Update the indications of each icon.
                        //This may not be the most optimized way to update icon indications.
                        //For Component indications, each component listens for the QnetClientThread message event
                        //and handles their own updates.
                        for (var i=0; i < activeIcons.length; i++)
                        {
                            UpdateIconStatus(activeIcons[i], msg.data.indication);
                        }
                       
                        //Since this event will be fired periodically, it is a convenient time to 
                        //request a panel file via the developer tool if the user has clicked on the open file button.
                        if (requestFileName)
                        {
                            $.get("http://localhost:9999/open/filename/", function(data){
                                if (data !== "")
                                {
                                    //success... open the panel file
                                    requestFileName = false;
                                    OpenPanel("http://localhost:9999/" + data);
                                }
                            }, "text");  
                        }    
                    });
                } 
                else
                {
                    //Hopefully this message will never be displayed when using the newer browsers.
                    alert("The QLCP soft panel will not work with this browser because it does not support web workers.");
                }
                
                //Request the control labels (status names) from the QLCP-NET Developer tool and create the control status name options.               
                var controlLabelsRequest = $.get("http://localhost:9999/control_labels/", function(data){
                    
                    //Success! So enter design mode and prepare the UI accordingly.
                    DesignMode();
                    
                    developerMode = true;
                    
                    //send a message to the web worker so that it knows the developer mode state and host url
                    QnetClientThread.postMessage({devMode:developerMode, url:hostUrl, sendControl:false, bitNbr:0, bitState:0, sendAllControls:false, cntrlBits:controlBits});
                    
                    //update the control labels
                    UpdateControlLabels(data);
                    
                    //Request the indication labels from the QLCP GUI Developer tool and create the indication status name options.
                    $.get("http://localhost:9999/indication_labels/", function(data){
                        
                        UpdateIndicationLabels(data);
                        
                    }, "text");
                    
                }, "text");     
                
                //If there is no response from the developer tool, assume this is a live connection to the QLCP itself.
                //Request the location information and then the panel file from the QLCP server.
                controlLabelsRequest.fail(function() { 

                    //Request the location information from the QLCP       
                    $.get("http://" + hostUrl + ":54321/info/", function(data){
                        
                        //Success! Received the location information. Now, display it in the message dialog box.
                        $( "#message_dialog" ).dialog( "option", "title", "Control Panel Request" );

                        document.getElementById("message_dialog").innerHTML = '<div><img style="float: left; margin-top: 5px; width: 36px" src="Images/warning_icon.png"/>'
                            + '<div style="float: left; margin-left: 15px; margin-top: 6px; width: 350px">You have requested the control panel for the following location:<br/><br/>'
                            + 'IP Address:&nbsp&nbsp' + hostUrl + '<br/>'
                            + 'Controller Name:&nbsp&nbsp' + data.split(",")[0] + '<br/>'
                            + 'Location Name:&nbsp&nbsp' + data.split(",")[1] + '<br/>'
                            + 'Railroad Name:&nbsp&nbsp' + data.split(",")[2] + '<br/>'
                            + 'Subdivision Name:&nbsp&nbsp' + data.split(",")[3] + '<br/>'
                            + '</div></div>';   
                
                        ShowMessageDialog();       
                    
                    }, "text");
  
                    //request and open the panel file
                    OpenPanel("http://" + hostUrl + "/qlcp.cpf");      
                        
                    //send a message to the web worker so that it knows the developer mode state and host url
                    QnetClientThread.postMessage({devMode:developerMode, url:hostUrl, sendControl:false, bitNbr:0, bitState:0, sendAllControls:false, cntrlBits:controlBits});

                    //enter run mode
                    RunMode();

                    //hide the design mode button
                    document.getElementById("design_button").style.visibility = "hidden";
                });
                     
                //this will periodically call UpdateComponentLayer to draw components as they are added/updated.
                window.requestAnimationFrame(UpdateComponentLayer);
                
                //move the component layer in front of the icons layer using zIndex
                document.getElementById("component_container").style.zIndex = 10;
                
                //prevent context menu from popping up when user clicks on right mouse button
                document.addEventListener("contextmenu", function(e){
                    e.preventDefault();
                }, false);
                
            }); 
            
            function UpdateControlLabels(data)
            {
                var controlLabels = data.split(",");
                controlOptionsStr = "<option>None</option>";

                //color-code groups of status names (128 per) to make it easier for multiple-server soft panel designs
                var listBackgroundColor = "#c6e0b4";
                for(var i=0; i < controlLabels.length - 1; i++)
                {
                    if (i === 128) listBackgroundColor = "#ffe699";
                    if (i === 256) listBackgroundColor = "#bdd7ee";
                    if (i === 384) listBackgroundColor = "#f8cbad";

                    controlOptionsStr = controlOptionsStr + "<option style='background-color: " + listBackgroundColor + "'>" + controlLabels[i] + "</option>";
                    controlBits[i] = 0; //initialize control bit state
                }
            }
            
            function UpdateIndicationLabels(data)
            {
                var indicationLabels = data.split(",");
                indicationOptionsStr = "<option>None</option>";
                listBackgroundColor = "#c6e0b4";

                //color-code groups of status names (128 per) to make it easier for multiple-server soft panel designs
                for(var i=0; i < indicationLabels.length - 1; i++)
                {
                    if (i === 128) listBackgroundColor = "#ffe699";
                    if (i === 256) listBackgroundColor = "#bdd7ee";
                    if (i === 384) listBackgroundColor = "#f8cbad";
                    indicationOptionsStr = indicationOptionsStr + "<option style='background-color: " + listBackgroundColor + "'>" + indicationLabels[i] + "</option>";            
                }             
            }
           
            //This function is periodically called using requestAnimationFrame to draw the component
            //layer. This is required when changes are made to the appearance of a component.
            //Since design mode will typically be used on a fast enough laptop PC, the component layer
            //is drawn as frequently as the browser can. For run mode, the redrawComponents flag is used
            //to only draw the layer when necessary.
            function UpdateComponentLayer()
            {                   
                if (designMode)
                {
                    component_layer.draw();
                }
                else if (redrawComponents)
                {
                    component_layer.draw();
                    redrawComponents = false;
                }
                               
                //requestAnimationFrame works using recursion
                window.requestAnimationFrame(UpdateComponentLayer);
            }

            function InitGrid()
            {              
                //Create the divs for the drawing grid and the grid lines
                var htmlGrid = "";
                var htmlGridLines = "";
                var rowIndex = 0;
                var colIndex = 0;
                
                var i;
                var j;
                for (i=0; i < nbrRows; i++)
                {                   
                    for(j=0; j < nbrCols; j++)
                    {
                        htmlGrid = htmlGrid + '<div id="' + rowIndex + ',' + colIndex + '" class="gridCell" style="float: left" data-idiv="" '
                            + 'ondrop="drop(event)" ondragover="allowDrop(event)" onmousedown="GetSelectedGridCell(this)" onmouseup="FindSelectedElements(this)"></div>';
                        htmlGridLines = htmlGridLines + '<div id="s' + rowIndex + ',' + colIndex + '" class="gridLines" style="float: left"></div>';
                        colIndex++;
                    } 
                    //start the next row
                    rowIndex++;
                    colIndex = 0;
                    htmlGrid = htmlGrid + '<div id="' + rowIndex + ',' + colIndex + '" class="gridCell" data-idiv="" '
                        + 'ondrop="drop(event)" ondragover="allowDrop(event)" onmousedown="GetSelectedGridCell(this)" onmouseup="FindSelectedElements(this)"></div>';
                    htmlGridLines = htmlGridLines + '<div id="s' + rowIndex + ',' + colIndex + '" class="gridLines"></div>';
                }
                             
                document.getElementById("grid").innerHTML = htmlGrid;
                document.getElementById("grid_lines").innerHTML = htmlGridLines;
                
                //put the trash can in the top left grid cell and make it so that drops can only occur on the trash can and not the grid cell div
                var gridId = "0,0";

                document.getElementById(gridId).ondrop = "";
                document.getElementById(gridId).ondragover = "";
                document.getElementById(gridId).onmousedown = "";
                document.getElementById(gridId).onmouseup = "";
                document.getElementById(gridId).draggable = "false";
                document.getElementById(gridId).innerHTML = '<img id="trash_can" data-idiv="" src="Images/trash.png" draggable="false" '
                    + 'style="margin-left: 7px; margin-top: 7px" ondrop="drop(event)" ondragover="allowDrop(event)"/>';                    
                                    
                if (designMode === false)
                {
                    //Hide grid borders
                    document.getElementById("grid_lines").style.visibility = "hidden";
 
                    //Hide the trash can
                    document.getElementById("trash_can").style.visibility = "hidden";    
                }
            }
            
            function InitDialogBoxes()
            {
                $( "#icons_dialog" ).dialog({ autoOpen: false });
                $( "#icons_dialog" ).dialog({ resizable: false });
                $( "#icons_dialog" ).dialog({ width: 241 });
                $( "#icons_dialog" ).dialog({ appendTo: "#buttonBackground" });
                    
                $( "#properties_dialog" ).dialog({ autoOpen: false });                
                $( "#properties_dialog" ).dialog({ minWidth: "500" });
                $( "#properties_dialog" ).dialog({ height: "700" });
                $( "#properties_dialog" ).dialog({ resizable: true });
                $( "#properties_dialog" ).dialog({ appendTo: "#buttonBackground" });
                $( "#properties_dialog" ).dialog({ close: ClearPreviousSelection });
                
                $( "#controls_dialog" ).dialog({ autoOpen: false });
                $( "#controls_dialog" ).dialog({ width: "auto" });
                $( "#controls_dialog" ).dialog({ height: "auto" });
                $( "#controls_dialog" ).dialog({ resizable: false });
                $( "#controls_dialog" ).dialog({ appendTo: "#headerWrapper" });
                
                $( "#clear_all_dialog" ).dialog({ autoOpen: false });
                $( "#clear_all_dialog" ).dialog({ width: "auto" });
                $( "#clear_all_dialog" ).dialog({ height: "auto" });
                $( "#clear_all_dialog" ).dialog({ resizable: false });
                $( "#clear_all_dialog" ).dialog({ modal: true });
                $( "#clear_all_dialog" ).dialog({ appendTo: "#buttonBackground" });
                                             
                $( "#message_dialog" ).dialog({ autoOpen: false });
                $( "#message_dialog" ).dialog({ width: "auto" });
                $( "#message_dialog" ).dialog({ height: "auto" });
                $( "#message_dialog" ).dialog({ resizable: false });
                $( "#message_dialog" ).dialog({ modal: true });
                $( "#message_dialog" ).dialog({ appendTo: "#headerWrapper" });        
                                
                $( "#passcode_dialog" ).dialog({ autoOpen: false });
                $( "#passcode_dialog" ).dialog({ width: "auto" });
                $( "#passcode_dialog" ).dialog({ height: "auto" });
                $( "#passcode_dialog" ).dialog({ resizable: false });
                $( "#passcode_dialog" ).dialog({ modal: true });
                $( "#passcode_dialog" ).dialog({ appendTo: "#headerWrapper" });
            }
            
            function CloseAllDialogBoxes()
            {
                if ($( "#icons_dialog" ).dialog( "isOpen" ))
                {
                    $( "#icons_dialog" ).dialog( "close" );
                }

                if ($( "#properties_dialog" ).dialog( "isOpen" ))
                {
                    $( "#properties_dialog" ).dialog( "close" );
                }
                
                if ($( "#controls_dialog" ).dialog( "isOpen" ))
                {
                    $( "#controls_dialog" ).dialog( "close" );
                }
                                                
                if ($( "#clear_all_dialog" ).dialog( "isOpen" ))
                {
                    $( "#clear_all_dialog" ).dialog( "close" );
                }
                                                
                if ($( "#message_dialog" ).dialog( "isOpen" ))
                {
                    $( "#message_dialog" ).dialog( "close" );
                }
                                                                
                if ($( "#passcode_dialog" ).dialog( "isOpen" ))
                {
                    $( "#passcode_dialog" ).dialog( "close" );
                }
            }
            
            function DesignMode()
            {
                designMode = true;
                
                //show button background
                document.getElementById("buttonBackground").style.visibility = "visible"; 

                //hide run mode buttons
                document.getElementById("fullscreen_button").style.visibility = "hidden";
                document.getElementById("design_button").style.visibility = "hidden";
                document.getElementById("send_all_button").style.visibility = "hidden";
                document.getElementById("clear_controls_button").style.visibility = "hidden";
                    
                //show design mode buttons
                document.getElementById("designButtons").style.visibility = "visible";
                document.getElementById("clear_all_button").style.visibility = "visible";
                
                SetPanelStyle(false);

                //Show grid borders
                document.getElementById("grid_lines").style.visibility = "visible";
                
                //close dialog boxes if they are open
                CloseAllDialogBoxes();
                
                //turn off all latched switches
                var components = component_layer.getChildren();
                var clickEvent = document.createEvent('MouseEvents');
                clickEvent.initEvent('mousedown', true, true);
                
                //set flag to tell the component event handlers that the events we are about to fire are for initialization
                initLatchedSwitches = true;
                
                for (var i = 0; i < components.length; i++)
                {
                    var comp = components[i];

                    if (typeof comp.attrs.type !== 'undefined')
                    {
                        if ((comp.attrs.type === 'togSw' || comp.attrs.type === 'keyswitch') && comp.attrs.state === true)
                        {
                            //turn it off by firing a click event to toggle the state
                            comp.dispatchEvent(clickEvent);
                        }
                    }
                }
                
                initLatchedSwitches = false;
                
                //clear selection of the last latched switch that may have been selected by initialization sequence
                ClearPreviousSelection();               
                                
                //clear any selected icons                
                ClearSelection();
                
                //change styles to maximize stage and remove scrollable div area
                document.getElementsByTagName("body")[0].style.overflow = "visible";
                document.getElementById("scroll_container").style.top = "93px";
                document.getElementById("scroll_container").style.width = "16020px";
                document.getElementById("scroll_container").style.height = "10020px";
            }
            
            function RunMode()
            {   
                designMode = false;
                
                //hide design mode buttons
                document.getElementById("designButtons").style.visibility = "hidden";
                
                //hide tools buttons
                HideToolsButtons();

                //show run mode buttons
                document.getElementById("fullscreen_button").style.visibility = "visible";
                document.getElementById("design_button").style.visibility = "visible";

                //If component style panel, hide button menu and run buttons
                if (componentStyle)
                {
                    document.getElementById("send_all_button").style.visibility = "hidden";
                    document.getElementById("clear_controls_button").style.visibility = "hidden";
                }
                else
                {
                    document.getElementById("send_all_button").style.visibility = "visible";
                    document.getElementById("clear_controls_button").style.visibility = "visible";
                }
                    
                //Hide grid borders
                document.getElementById("grid_lines").style.visibility = "hidden";
                
                //Hide the trash can
                document.getElementById("trash_can").style.visibility = "hidden";
                
                //hide any dialog boxes from design mode if they are open
                CloseAllDialogBoxes();
                
                //clear any selected icons                
                ClearSelection();
                
                ClearSelectedComponents();

                //deselect any buttons for dropping components
                DeselectDropComponentButtons();
                
                //Initialize control bit states and send all controls...
                //First, clear all states in the controlBits array.
                var i=0
                for (i=0; i < controlBits.length; i++)
                {
                    controlBits[i] = 0;
                }

                //Now, turn off all toggle switches and set their off states to 1.
                var components = component_layer.getChildren();
                var clickEvent = document.createEvent('MouseEvents');
                clickEvent.initEvent('mousedown', true, true);  

                //set flag to tell the component event handlers that the events we are about to fire are for initialization
                initLatchedSwitches = true;

                for (i = 0; i < components.length; i++)
                {
                    var comp = components[i];

                    //if component should be locked, lock it
                    if (typeof comp.attrs.unlocked !== 'undefined')
                    {
                        comp.attrs.unlocked = false;
                    }

                    if (typeof comp.attrs.type !== 'undefined')
                    {
                        if (comp.attrs.type === 'togSw')
                        {
                            comp.dispatchEvent(clickEvent); //turn switch on
                            comp.dispatchEvent(clickEvent); //turn switch off
                        }
                        if (comp.attrs.type === 'keyswitch')
                        {
                            comp.dispatchEvent(clickEvent); //set to position 1 (up)
                        }
                    }
                }

                initLatchedSwitches = false;

                //change styles to minimize stage in a scrollable area
                document.getElementsByTagName("body")[0].style.overflow = "hidden";
                document.getElementById("scroll_container").style.top = "50px";
                document.getElementById("scroll_container").style.width = "calc(100% - 2px)";
                document.getElementById("scroll_container").style.height = "calc(100vh - 82px)";
                document.getElementById("buttonBackground").style.visibility = "hidden"; 
                
                SendAllControls();
            }
                     
            function SetPanelStyle(userChangedSelection)
            {
                //If the page is not clear of components, warn the user that they will be deleted and allow the user to proceed or cancel action.
                if (userChangedSelection && ((component_layer.getChildren().length > 0) || (activeIcons.length > 0)))
                {                                                                
                    //return selected index to other style
                    if(document.getElementById("design_style_option").selectedIndex === 0)
                    {
                       document.getElementById("design_style_option").selectedIndex = 1
                    }
                    else
                    {
                       document.getElementById("design_style_option").selectedIndex = 0
                    }         

                    $( "#message_dialog" ).dialog( "option", "title", "Change Design Style" );

                    document.getElementById("message_dialog").innerHTML = '<div><img style="float: left; margin-top: 5px; width: 36px" src="Images/warning_icon.png"/>'
                        + '<div style="float: left; margin-left: 15px; margin-top: 6px; width: 220px">This will delete everything in your current design!</div></div>';   

                    $( "#message_dialog" ).dialog( "option", "buttons", [
                        { 
                            text: "OK", click: function() { 
                                $( this ).dialog( "close" );
                                ClearAllComponents();
                                
                                //change selected index for panel style
                                if(document.getElementById("design_style_option").selectedIndex === 0)
                                {
                                   document.getElementById("design_style_option").selectedIndex = 1
                                }
                                else
                                {
                                   document.getElementById("design_style_option").selectedIndex = 0
                                }
                                SetPanelStyle(false);
                            }  
                        },
                        { 
                            text: "Cancel", click: function() { 
                                
                                $( this ).dialog( "close" ); 
                            }  
                        }
                    ]);
                    
                    $( "#message_dialog" ).dialog( "open" );        
                }
                else
                {
                    if (document.getElementById("design_style_option").selectedIndex === 1)
                    {
                        //set tools up for Mosaic style
                        ShowMosaicStyleTools();
                        document.getElementById("component_container").style.zIndex = 3;
                        componentStyle = false;
                    }         
                    else
                    {   
                        //set tools up for Components style
                        ShowHardPanelTools();
                        document.getElementById("component_container").style.zIndex = 10;
                        componentStyle = true;
                    }
                }
                               
                //deselect any buttons for dropping components
                DeselectDropComponentButtons();
                
                //clear any selected icons                
                ClearSelection();
                
                //hide any dialog boxes from design mode if they are open
                if ($( "#icons_dialog" ).dialog( "isOpen" ))
                {
                    $( "#icons_dialog" ).dialog( "close" );
                }
                if ($( "#properties_dialog" ).dialog( "isOpen" ))
                {
                    $( "#properties_dialog" ).dialog( "close" );
                }
                                               
                //Hide or show the trash can based on design style
                if (componentStyle)
                {
                    document.getElementById("trash_can").style.visibility = "hidden";  
                }
                else
                {
                    document.getElementById("trash_can").style.visibility = "visible";  
                }              
            }
            
            function ShowMosaicStyleTools()
            {
                HideToolsButtons();
                document.getElementById("icons_button").style.visibility = "visible";
                document.getElementById("text_button").style.visibility = "visible";
                document.getElementById("clear_all_button").style.visibility = "visible";
                document.getElementById("properties_button").style.visibility = "visible";
            }
            
            //Hard panels is the old name for Component style.
            function ShowHardPanelTools()
            {
                HideToolsButtons();
                document.getElementById("text_button").style.visibility = "visible";
                document.getElementById("componentButtons").style.visibility = "visible";
                document.getElementById("increase_button").style.visibility = "visible";
                document.getElementById("decrease_button").style.visibility = "visible";
                document.getElementById("remove_background_button").style.visibility = "visible";
                document.getElementById("properties_button").style.visibility = "visible";
                document.getElementById("clear_all_button").style.visibility = "visible";
                document.getElementById("position_display").style.visibility = "visible";
            }
            
            function HideToolsButtons()
            {
                document.getElementById("icons_button").style.visibility = "hidden";
                document.getElementById("text_button").style.visibility = "hidden";
                document.getElementById("componentButtons").style.visibility = "hidden";
                document.getElementById("increase_button").style.visibility = "hidden";
                document.getElementById("decrease_button").style.visibility = "hidden";
                document.getElementById("remove_background_button").style.visibility = "hidden";
                document.getElementById("properties_button").style.visibility = "hidden";
                document.getElementById("clear_all_button").style.visibility = "hidden";
                document.getElementById("position_display").style.visibility = "hidden";
            }
            
            function ClearSelection()
            {
                //clear selection for all active icons
                for(var i=0; i < activeIcons.length; i++)
                {
                    document.getElementById(activeIcons[i]).style.border = "none";
                    
                    //Put margins back to zero...
                    //When border is shown for selected icon, margins have to be adjusted to negative values,
                    //otherwise image is shifted.
                    document.getElementById(activeIcons[i]).style.marginTop = "0px";
                    document.getElementById(activeIcons[i]).style.marginLeft = "0px";
                }   
            }
            
            function GetSelectedGridCell(grid_cell)
            {
                //This function is called on a mouse down event over the grid area.
                //The id of the first grid cell for the selection area is saved.
                selectionStart = grid_cell.id;
            }
            
            function FindSelectedElements(grid_cell)
            {
                //This function is called on a mouse up event over the grid area.
                //It determines the IDs of all icons in the selected area.
                //This function is intended to support multiple selection
                //for future use. At this time, dragging multiple icons is not
                //supported however.
                              
                var selectionEnd = grid_cell.id;
                
                //Determine the coordinates of the selection area.
                //TO-DO: support selection if motion is top right to bottom left.
                //Currently supports a top-left to bottom-right selection motion only.
                var firstRow = selectionStart.split(",")[0];
                var lastRow = selectionEnd.split(",")[0];
                var firstCol = selectionStart.split(",")[1];
                var lastCol = selectionEnd.split(",")[1];
                var foundIcon = false;
                
                //Clear any other icons that may currently be selected
                ClearSelection();
                
                //clear selected icons array
                selectedIcons.length = 0;
                
                //Find the selected icons and show them as such
                for (var row=firstRow; row <= lastRow; row++)
                {
                    for (var col=firstCol; col <= lastCol; col++)
                    {
                        var gridId = row + "," + col;
                        var gridObj = document.getElementById(gridId);
                        
                        //get icon in this grid cell if there is one
                        if (gridObj.children[0])
                        {
                            ShowSelected(gridObj.children[0].id);
                            selectedIcons.push(gridObj.children[0].id);
                            foundIcon = true;
                        }
                    }
                }
               
                //If there are no icons in the selection area, close the properties dialog
                if (false == foundIcon)
                {
                    $( "#properties_dialog" ).dialog( "close" );
                }                
            }
            
            function ShowSelected(id)
            {                  
                //Show the icon as selected (the negative margins compensate for the shifting when a 1 pixel border is added). 
                //The margins will be set back to zero when deselected.
                document.getElementById(id).style.border = '#2593B8 1px solid';   
                document.getElementById(id).style.marginTop = '-1px';   
                document.getElementById(id).style.marginLeft = '-1px';  
            }
            
            function ShowIconsDialog()
            {                   
                //move the component layer behind the icons layer
                document.getElementById("component_container").style.zIndex = 3;
                
                //Create the icons and display the icons dialog box.                    
                var htmlStr = "";
                              
                for(var i=0; i < iconTypes.length; i++)
                {
                    idNbr++;

                    htmlStr = htmlStr + '<div id="d'+ i + '" class="iconDialogDiv">'
                        + '<canvas id="g' + idNbr + '" title="" '
                        + 'data-idiv="d' + i + '" '
                        + 'draggable="true" ondragstart="drag(event)" '
                        //+ 'ondblclick="UpdateIconStatus(' + '\'g' + idNbr + '\')" '
                        + 'onmouseup="HandleClick(' + '\'g' + idNbr + '\')" '
                        + 'width="48" height="48"' + customDataStr + '></canvas></div>';
                }
                
                document.getElementById("icons_dialog").innerHTML = htmlStr;
                    
                DrawIconImages();
                    
                //Open the icons dialog
                $( "#icons_dialog" ).dialog( "open" );
                
                //Clear any selected text
                ClearPreviousSelection();
                
                //Clear any selected icons
                ClearSelection();
                
                DeselectDropComponentButtons();
                $( '#icons_button' ).addClass("buttonDown");
                dropMode['icons'] = true;
                
                if ($( "#properties_dialog" ).dialog( "isOpen" ))
                {
                    $( "#properties_dialog" ).dialog( "close" );
                }
            }
            
            function allowDrop(ev)
            {
                //Icons can only be dropped if in Design Mode
                //This overrides the default behavior, which is to prevent drops
                if (designMode)
                {
                    ev.preventDefault();
                }
            }

            function drag(ev)
            {
                //set data to be transferred to the target element
                ev.dataTransfer.setData("src_div",ev.target.dataset.idiv);
                ev.dataTransfer.setData("src_id",ev.target.id);
            }


            function drop(ev)
            {                       
                var div=ev.dataTransfer.getData("src_div"); //get the div of the dropped icon
                var id=ev.dataTransfer.getData("src_id"); //get the id of the dropped icon  
                
                //Check the data-idiv attribute to determine if the drop target is a blank div or one that already
                //contains an icon. If data-idiv contains a value, the grid cell contains an icon and 
                //the drop should not be allowed.
                //Also, do not allow a drop if this is an icon that is pulled from the icons dialog
                //directly to the trash can (someone will try this I'm sure)
                if ("" != ev.target.dataset.idiv | "trash_can" == ev.target.id & 1 == document.getElementById(id).dataset.newdrop)
                {
                    return;
                }
                
                ev.preventDefault(); //this allows the drop which is normally prevented as default
               
                var debugStr = "Dropped: " + id + "<br>";
              
                //Complete the drop by appending it as a child of the target div
                ev.target.appendChild(document.getElementById(id));
                                
                //If the drop target is the trash can, delete the icon ID from the activeIcons array and exit
                if ("trash_can" == ev.target.id)
                {
                    DeleteIcon(id);

                    //Close the properties dialog box if open because it could be open for the deleted icon
                    if ($( "#properties_dialog" ).dialog( "isOpen" ))
                    {
                        $( "#properties_dialog" ).dialog( "close" );
                    }
                    
                    return;                        
                }
                
                //set the gridlocation attribute of the dropped icon
                document.getElementById(id).dataset.gridlocation = ev.target.id;

                //Add the icon ID to the active list if this is a new drop (from icons dialog to grid).
                if (1 == document.getElementById(id).dataset.newdrop)
                {
                    activeIcons.push(id);
                    document.getElementById(id).dataset.newdrop = 0;
                }

                debugStr = debugStr + "Active Icons: "
                for (var i=0; i < activeIcons.length; i++)
                {
                    debugStr = debugStr + activeIcons[i] + ", ";
                }
                debugStr = debugStr + "<br>";
                
                //console.log(debugStr);
                                   
                idNbr++; //increment the global ID number
                
                //Replace the icon that was removed from the icon dialog with a new icon and ID
                document.getElementById(div).innerHTML = '<canvas id="g' + idNbr + '" title="" '
                    + 'data-idiv="' + div + '" '
                    + 'draggable="true" ondragstart="drag(event)" '
                    //+ 'ondblclick="UpdateIconStatus(' + '\'g' + idNbr + '\')" '
                    + 'onmouseup="HandleClick(' + '\'g' + idNbr + '\')" '
                    + 'width="48" height="48"' + customDataStr + '></canvas>';  

                DrawIcon('g' + idNbr, 0, 0, defaultTrackColor, defaultTrackColor);

            }
            
            function DeleteIcon(id)
            {
                //Shift (off the top) icon IDs from the activeIcons array and check if it's the one to delete.
                //If not, push it back on the array until the correct ID is located, then return.
                for (var i=0; i < activeIcons.length; i++)
                {
                    var iconId = activeIcons.shift();

                    if (id == iconId)
                    {
                        //got it, now return
                        return;
                    }
                    else
                    {
                        //that's not it so put it back
                        activeIcons.push(iconId);
                    }
                }
            }

            function HandleClick(id)
            {       
                //alert(document.getElementById(id).dataset.gridlocation);
                //console.log("Enter handle click");
                
                //If this icon is in the icons dialog (hasn't been dropped yet), return
                if (1 == document.getElementById(id).dataset.newdrop)
                {
                    return;
                }
                
                //Based on the icon's type, which is determined by its original div location
                //in the icons dialog box, build an html string for its properties dialog box.
                //This string will set the innerHTML of the properties dialog box.
                var htmlStr;

                if (designMode)
                {
                    var hasControls;
                    var hasPositionIndications;
                    var hasMultipleColorIndications;
                    var posStr1 = "";
                    var posStr2 = "";
                    var colorStr1 = "";
                    var colorStr2 = "";
                    var cntrlPropStr = "";
                    var indPropStr1 = "";
                    var indPropStr2 = "";
                    var indPropStr3 = "";
                    
                    //First, find the icon type to determine how to build the properties dialog box
                    var div = document.getElementById(id).dataset.idiv;
                    var type = iconTypes[div.substr(1)];
                    
                    //Based on the icon type, set attributes to determine how to create the properties string.

                    switch(type)
                    {
                        case "track":
                        case "track_90":
                        case "track_45":
                        case "track_135":
                        case "track_corner":
                        case "track_corner_flip_v":
                        case "track_corner_flip_h":
                        case "track_corner_180":
                        case "track_corner_90_flip_h":
                        case "track_corner_90":                        
                        case "track_corner_270":
                        case "track_corner_270_flip_h":
                         
                            hasControls = true;
                            hasPositionIndications = false;
                            hasMultipleColorIndications = false;
                            colorStr1 = "Track";
                          
                            break;
                        
                        case "track_cross":             
                        case "track_cross_diag":
                            
                            hasControls = true;
                            hasPositionIndications = false;
                            hasMultipleColorIndications = true;
                            colorStr1 = "Main Track";
                            colorStr2 = "Crossing Track";
                                             
                            break;

                        case "switch":
                        case "switch_90":
                        case "switch_180":
                        case "switch_270":
                        case "switch_flip_h":
                        case "switch_flip_v":
                        case "switch_90_flip_h":
                        case "switch_270_flip_h":  
                            
                            hasControls = true;
                            hasPositionIndications = true;
                            hasMultipleColorIndications = false;
                            posStr1 = "Normal Position";
                            posStr2 = "Reverse Position";
                            colorStr1 = "Track";

                            break;

                        case "signal":
                        case "signal_90":
                        case "signal_180":
                        case "signal_270":
                        case "signal_flip_h":
                        case "signal_flip_v":
                        case "signal_90_flip_h":
                        case "signal_270_flip_h":
                        case "signal_45":
                        case "signal_135":
                        case "signal_225":
                        case "signal_315":
                        case "signal_45_flip_h":
                        case "signal_135_flip_h":
                        case "signal_225_flip_h":
                        case "signal_315_flip_h":
                            
                            hasControls = true;
                            hasPositionIndications = false;
                            hasMultipleColorIndications = true;
                            colorStr1 = "Signal";
                            colorStr2 = "Track";
                            break;
                            
                        case "arrow":
                        case "arrow_90":
                        case "arrow_180":
                        case "arrow_270":
                        case "circle":
                           
                            hasControls = true;
                            hasPositionIndications = false;
                            hasMultipleColorIndications = false;
                            colorStr1 = "";
                            break;
                        
                        default:                       
                    } 
                
                    //Now, build the html string to create the properties dialog for the selected icon
                    //If this icon has controls, add the labels and drop down lists to select status names. Also, add the labels and text boxes to enter custom button names.
                    if (hasControls)
                    {
                        for (var i=0; i < nbrControls ; i++)
                        {
                            var labelNbr = i + 1;
                            cntrlPropStr = cntrlPropStr + '<p style="margin-top: 10px">Control Status ' + labelNbr + ':&nbsp&nbsp'
                                + '<td><select id="control' + i + '_label_options" onchange="SetControlBitProperty(\'' + id + '\',\'control' + i + '\')" style="float: right; width: 210px">' 
                                + controlOptionsStr
                                + '</select></td></p>'
                                + '<p style="margin-top: 10px">Button Label ' + labelNbr + ':&nbsp&nbsp'
                                + '<input id="control' + i + '_button_label" type="text" oninput="SetControlNameProperty(\'' + id + '\',\'control' + i + '\')" style="float: right; width: 210px"/></p>';
                        }
                    }
                
                    //If this icon has position indications, add the labels and drop down lists to select status names.
                    if (hasPositionIndications)
                    {
                        indPropStr1 = '<p style="margin-top: 10px">' + posStr1 + ' Indication:&nbsp&nbsp'
                            + '<td><select id="indication0_state_options" onchange="SetIndicationProperty(\'' + id + '\',\'indication0\')" style="float: right; width: 60px">' 
                            + '<option>True</option>'
                            + '<option>False</option>'
                            + '</select></td>'
                            + '<td><select id="indication0_label_options" onchange="SetIndicationProperty(\'' + id + '\',\'indication0\')" style="float: right; width: 150px">' 
                            + indicationOptionsStr
                            + '</select></td></p>'
                            + '<p style="margin-top: 10px">' + posStr2 + ' Indication:&nbsp&nbsp'
                            + '<td><select id="indication1_state_options" onchange="SetIndicationProperty(\'' + id + '\',\'indication1\')" style="float: right; width: 60px">' 
                            + '<option>True</option>'
                            + '<option>False</option>'
                            + '</select></td>'
                            + '<td><select id="indication1_label_options" onchange="SetIndicationProperty(\'' + id + '\',\'indication1\')" style="float: right; width: 150px">' 
                            + indicationOptionsStr
                            + '</select></td></p>';
                    }

                    //Add the labels and drop down lists to select status names for the first colored indication. Also add the lists to select the state option.
                    //The state option determines whether the color is displayed when the bit state is true or when it is false.
                    for (var i=2, k=0; i < iconColors.length + 2; i++, k++)
                    {
                        indPropStr2 = indPropStr2 + '<p style="margin-top: 10px">' 
                            + colorStr1 + ' Indication (' + iconColors[k] + '):&nbsp&nbsp'
                            + '<td><select id="indication'+ i + '_state_options" onchange="SetIndicationProperty(\'' + id + '\',\'indication' + i + '\')" style="float: right; width: 60px">' 
                            + '<option>True</option>'
                            + '<option>False</option>'
                            + '</select></td>'
                            + '<td><select id="indication'+ i + '_label_options" onchange="SetIndicationProperty(\'' + id + '\',\'indication' + i + '\')" style="float: right; width: 150px">' 
                            + indicationOptionsStr
                            + '</select></td></p>';
                    }

                    //If this icon has multiple colored indications, add the labels and drop down lists to select status names.
                    if (hasMultipleColorIndications)
                    {
                        for (var i=iconColors.length + 2, k=0; i < iconColors.length * 2 + 2; i++, k++)
                        {
                            indPropStr3 = indPropStr3 + '<p style="margin-top: 10px">' 
                                + colorStr2 + ' Indication (' + iconColors[k] + '):&nbsp&nbsp'
                                + '<td><select id="indication'+ i + '_state_options" onchange="SetIndicationProperty(\'' + id + '\',\'indication' + i + '\')" style="float: right; width: 60px">' 
                                + '<option>True</option>'
                                + '<option>False</option>'
                                + '</select></td>'
                                + '<td><select id="indication'+ i + '_label_options" onchange="SetIndicationProperty(\'' + id + '\',\'indication' + i + '\')" style="float: right; width: 150px">' 
                                + indicationOptionsStr
                                + '</select></td></p>';
                        }
                    }
                    
                    //Build the final string and assign it to the properties dialog box
                    htmlStr = cntrlPropStr + indPropStr1 + indPropStr2 + indPropStr3;           
                    document.getElementById("properties_dialog").innerHTML = htmlStr;

                    //get the current property values for this icon
                    if (hasControls)
                    {
                        document.getElementById("control0_button_label").value = document.getElementById(id).dataset["control0"].split(",")[0];    
                        document.getElementById("control0_label_options").selectedIndex = document.getElementById(id).dataset["control0"].split(",")[1];
                        document.getElementById("control1_button_label").value = document.getElementById(id).dataset["control1"].split(",")[0];    
                        document.getElementById("control1_label_options").selectedIndex = document.getElementById(id).dataset["control1"].split(",")[1];                   
                        document.getElementById("control2_button_label").value = document.getElementById(id).dataset["control2"].split(",")[0];    
                        document.getElementById("control2_label_options").selectedIndex = document.getElementById(id).dataset["control2"].split(",")[1];                   
                        document.getElementById("control3_button_label").value = document.getElementById(id).dataset["control3"].split(",")[0];    
                        document.getElementById("control3_label_options").selectedIndex = document.getElementById(id).dataset["control3"].split(",")[1]; 
                        document.getElementById("control4_button_label").value = document.getElementById(id).dataset["control4"].split(",")[0];    
                        document.getElementById("control4_label_options").selectedIndex = document.getElementById(id).dataset["control4"].split(",")[1];             
                        document.getElementById("control5_button_label").value = document.getElementById(id).dataset["control5"].split(",")[0];    
                        document.getElementById("control5_label_options").selectedIndex = document.getElementById(id).dataset["control5"].split(",")[1];             
                        document.getElementById("control6_button_label").value = document.getElementById(id).dataset["control6"].split(",")[0];    
                        document.getElementById("control6_label_options").selectedIndex = document.getElementById(id).dataset["control6"].split(",")[1];             
                        document.getElementById("control7_button_label").value = document.getElementById(id).dataset["control7"].split(",")[0];    
                        document.getElementById("control7_label_options").selectedIndex = document.getElementById(id).dataset["control7"].split(",")[1];             
                        document.getElementById("control8_button_label").value = document.getElementById(id).dataset["control8"].split(",")[0];    
                        document.getElementById("control8_label_options").selectedIndex = document.getElementById(id).dataset["control8"].split(",")[1];             
                        document.getElementById("control9_button_label").value = document.getElementById(id).dataset["control9"].split(",")[0];    
                        document.getElementById("control9_label_options").selectedIndex = document.getElementById(id).dataset["control9"].split(",")[1];             
                    }  

                    if (hasPositionIndications)
                    {
                        document.getElementById("indication0_label_options").selectedIndex = document.getElementById(id).dataset["indication0"].split(",")[0];
                        document.getElementById("indication0_state_options").selectedIndex = document.getElementById(id).dataset["indication0"].split(",")[1];
                        document.getElementById("indication1_label_options").selectedIndex = document.getElementById(id).dataset["indication1"].split(",")[0];
                        document.getElementById("indication1_state_options").selectedIndex = document.getElementById(id).dataset["indication1"].split(",")[1];
                    }

                    for (var i=2; i < iconColors.length + 2; i++)
                    {
                        document.getElementById("indication" + i + "_label_options").selectedIndex = document.getElementById(id).dataset["indication" + i].split(",")[0];
                        document.getElementById("indication" + i + "_state_options").selectedIndex = document.getElementById(id).dataset["indication" + i].split(",")[1];
                    }

                    if (hasMultipleColorIndications)
                    {
                        for (var i=iconColors.length + 2; i < iconColors.length * 2 + 2; i++)
                        {
                            document.getElementById("indication" + i + "_label_options").selectedIndex = document.getElementById(id).dataset["indication" + i].split(",")[0];
                            document.getElementById("indication" + i + "_state_options").selectedIndex = document.getElementById(id).dataset["indication" + i].split(",")[1];
                        } 
                    }

                    //$( "#properties_dialog" ).dialog( "option", "title", "Properties" );
                      
                    //Open the properties dialog box                     
                    //$( "#properties_dialog" ).dialog( "open" );

                }
                else
                {
                    //We're not in Design Mode. We are currently in Run Mode so build the controls dialog box
                    htmlStr = GetControlsDialogContent(id);
                    
                    //If the returned string is empty, there are no user-defined controls for this icon so create a message indicating such.
                    if (htmlStr === "")
                    {
                        htmlStr = '<div><img style="float: left; margin-top: 5px; width: 36px" src="Images/info32px.png"/>'
                            + '<div style="float: left; margin-left: 15px; margin-top: 6px; width: 200px">There are no controls defined for this tile.</div></div>';;
                    }
                        
                    document.getElementById("controls_dialog").innerHTML = htmlStr;
                    
                    //Open the controls dialog box
                    $( "#controls_dialog" ).dialog( "open" );

                }
            }
            
            function SetControlBitProperty(id, property)
            {   
                //This function is called when a control status bit is changed to update the saved values.
                var buttonLabelId = property + "_button_label";
                var dropDownListId = property + "_label_options";
                
                if (document.getElementById(buttonLabelId).value == '')
                {
                    document.getElementById(buttonLabelId).value = document.getElementById(dropDownListId).value;  
                }
                document.getElementById(id).dataset[property] = document.getElementById(buttonLabelId).value + "," + document.getElementById(dropDownListId).selectedIndex;
            }   
            
            function SetControlNameProperty(id, property)
            {   
                //This function is called when a control status custom button label is changed to update the saved values.
                var buttonLabelId = property + "_button_label";
                var dropDownListId = property + "_label_options";
                document.getElementById(id).dataset[property] = document.getElementById(buttonLabelId).value + "," + document.getElementById(dropDownListId).selectedIndex;
            }      
            
            function SetIndicationProperty(id, property)
            {
                //This function is called when an indication status name or state option is changed to update the saved values.
                //The state option determines whether the color is displayed when the bit state is true or when it is false.
                var statusListId = property + "_label_options";
                var stateListId = property + "_state_options";
                document.getElementById(id).dataset[property] = document.getElementById(statusListId).selectedIndex + "," + document.getElementById(stateListId).selectedIndex;
            }
            
            function GetControlsDialogContent(id)
            {
                var htmlStr = "";
                var index;
                var bit;
                var buttonLabel;

                //For the controls that are being used (i.e. selected index of dropdown is > 0), create the dialog contents.
                for (var i=0; i < nbrControls; i++)
                {
                    index = document.getElementById(id).dataset["control" + i].split(",")[1];
                        
                    if (index > 0)
                    {
                        buttonLabel = document.getElementById(id).dataset["control" + i].split(",")[0];
                        bit = index - 1;
                        
                        var checkedState = "";
                        var disabledState = "";
                        if (1 == controlBits[bit])
                        {
                            checkedState = "checked";
                            disabledState = "disabled";
                        }
                          
                        htmlStr = htmlStr + '<input id="btn_' + i + '" type="button" style="width: 150px"' + disabledState + ' value="' + buttonLabel + '" onmousedown="SendControl(\'' + bit + '\',\'1\')" onmouseup="SendControl(\'' + bit + '\',\'0\')"/>'
                            + '<input type="checkbox" style="margin-left: 10px"' + checkedState + ' onchange="SetLatchedControl(\'' + bit + '\',Number(this.checked),\'' + i + '\')"/><br>'
                    }
                }
                return htmlStr;
            }          
          
            function DrawIconImages()
            {
                //This function draws the image for each icon type shown in the icons dialog box
                //The canvas element is used to draw each icon.
                for(var i=0; i < iconTypes.length; i++)
                {
                    var canvasId = document.getElementById('d' + i).getElementsByTagName("canvas")[0].id;
                    DrawIcon(canvasId, 0, 0, defaultTrackColor, defaultTrackColor);
                }               
            }
            
            //this function calls the function to draw the specific icon type
            function DrawIcon(canvasId, ind1, ind2, color1, color2)
            {
                var div = document.getElementById(canvasId).dataset.idiv;
                var type = iconTypes[div.substr(1)];
                
                switch(type)
                {
                    case "track":
                        DrawTrack(canvasId, color1, 0);
                        break;
                        
                    case "track_90":
                        DrawTrack(canvasId, color1, 90);
                        break;
                        
                    case "track_45":
                        DrawDiagonalTrack(canvasId, color1, 90);
                        break;
                        
                    case "track_135":
                        DrawDiagonalTrack(canvasId, color1, 0);                       
                        break;
                        
                    case "track_cross":
                        DrawCrossingTracks(canvasId, color1, color2);                     
                        break;
                                            
                    case "track_cross_diag":
                        DrawDiagonalCrossingTracks(canvasId, color1, color2);                     
                        break;
                        
                    case "track_corner":
                        DrawCornerTrack(canvasId, color1, 0);            
                        break;
                        
                    case "track_corner_flip_v":
                        DrawCornerTrack(canvasId, color1, 0, 'vertical');            
                        break;
                        
                    case "track_corner_flip_h":
                        DrawCornerTrack(canvasId, color1, 0, 'horizontal');            
                        break;
                        
                    case "track_corner_180":
                        DrawCornerTrack(canvasId, color1, 180);            
                        break;
                        
                    case "track_corner_90_flip_h":
                        DrawCornerTrack(canvasId, color1, 90, 'horizontal');            
                        break;
                        
                    case "track_corner_90":                        
                        DrawCornerTrack(canvasId, color1, 90);            
                        break;
                        
                    case "track_corner_270":
                        DrawCornerTrack(canvasId, color1, 270);            
                        break;
                                            
                    case "track_corner_270_flip_h":
                        DrawCornerTrack(canvasId, color1, 270, 'horizontal');            
                        break;
                        
                    case "switch":
                        DrawSwitch(canvasId, ind1, ind2, color1, 0);                     
                        break;
                        
                    case "switch_90":
                        DrawSwitch(canvasId, ind1, ind2, color1, 90);                     
                        break;
                        
                    case "switch_180":
                        DrawSwitch(canvasId, ind1, ind2, color1, 180);                     
                        break;
                        
                    case "switch_270":
                        DrawSwitch(canvasId, ind1, ind2, color1, 270);                     
                        break;
                        
                    case "switch_flip_h":
                        DrawSwitch(canvasId, ind1, ind2, color1, 0, 'horizontal');                     
                        break;
                        
                    case "switch_flip_v":
                        DrawSwitch(canvasId, ind1, ind2, color1, 0, 'vertical');                     
                        break;
                        
                    case "switch_90_flip_h":
                        DrawSwitch(canvasId, ind1, ind2, color1, 90, 'horizontal');                     
                        break;
                        
                    case "switch_270_flip_h":
                        DrawSwitch(canvasId, ind1, ind2, color1, 270, 'horizontal');                     
                        break;

                    case "signal":
                        DrawSignal(canvasId, color1, color2, 0);                     
                        break;
                        
                    case "signal_90":
                        DrawSignal(canvasId, color1, color2, 90);                     
                        break;
                        
                    case "signal_180":
                        DrawSignal(canvasId, color1, color2, 180);                     
                        break;
                        
                    case "signal_270":
                        DrawSignal(canvasId, color1, color2, 270);                     
                        break;
                        
                    case "signal_flip_h":
                        DrawSignal(canvasId, color1, color2, 0, 'horizontal');                     
                        break;
                        
                    case "signal_flip_v":
                        DrawSignal(canvasId, color1, color2, 0, 'vertical');                     
                        break;
                        
                    case "signal_90_flip_h":
                        DrawSignal(canvasId, color1, color2, 90, 'horizontal');                     
                        break;
                        
                    case "signal_270_flip_h":
                        DrawSignal(canvasId, color1, color2, 270, 'horizontal');                     
                        break;
                        
                    case "signal_45":
                        DrawSignal45(canvasId, color1, color2, 0);                     
                        break;
                        
                    case "signal_135":
                        DrawSignal45(canvasId, color1, color2, 90);                     
                        break;
                        
                    case "signal_225":
                        DrawSignal45(canvasId, color1, color2, 180);                     
                        break;
                        
                    case "signal_315":
                        DrawSignal45(canvasId, color1, color2, 270);                     
                        break;
                        
                    case "signal_45_flip_h":
                        DrawSignal45(canvasId, color1, color2, 0, 'horizontal');                     
                        break;
                        
                    case "signal_135_flip_h":
                        DrawSignal45(canvasId, color1, color2, 90, 'horizontal');                     
                        break;
                        
                    case "signal_225_flip_h":
                        DrawSignal45(canvasId, color1, color2, 180, 'horizontal');                     
                        break;
                        
                    case "signal_315_flip_h":
                        DrawSignal45(canvasId, color1, color2, 270, 'horizontal');                     
                        break;
                        
                    case "circle":
                        DrawCircle(canvasId, color1);                     
                        break;
                        
                    case "arrow":
                        DrawArrow(canvasId, color1, 0);                     
                        break;
                        
                    case "arrow_90":
                        DrawArrow(canvasId, color1, 90);                     
                        break;
                        
                    case "arrow_180":
                        DrawArrow(canvasId, color1, 180);                     
                        break;
                        
                    case "arrow_270":
                        DrawArrow(canvasId, color1, 270);                     
                        break;
                                                
                    default:                       
                }
            }
            
            function DrawTrack(canvasId, color, rotation)
            {               
                var canvas = document.getElementById(canvasId);
                var context = canvas.getContext('2d');

                //reset/clear the canvas
                canvas.width = canvas.width;
                canvas.height = canvas.height;
                
                var xoffset = canvas.width/2;
                var yoffset = canvas.height/2;
                context.translate(xoffset, yoffset);   
                
                //rotate context
                context.rotate(rotation*Math.PI/180);
                
                //draw track
                context.beginPath();
                context.moveTo(0-xoffset,24-yoffset);
                context.lineTo(48-xoffset,24-yoffset);
                context.lineWidth = 6;
                context.strokeStyle = color;
                context.stroke();    
            }
            
            function DrawDiagonalTrack(canvasId, color, rotation)
            {               
                var canvas = document.getElementById(canvasId);
                var context = canvas.getContext('2d');
                
                //reset/clear the canvas
                canvas.width = canvas.width;
                canvas.height = canvas.height;
                
                //translate context to center of canvas to support rotation and flipping
                var xoffset = canvas.width/2;
                var yoffset = canvas.height/2;
                context.translate(xoffset, yoffset);
                
                //rotate context
                context.rotate(rotation*Math.PI/180);

                //draw track
                context.beginPath();
                context.moveTo(0-xoffset,0-yoffset);
                context.lineTo(48-xoffset,48-yoffset);
                context.lineWidth = 6;
                context.strokeStyle = color;
                context.stroke();    
            }

            function DrawCrossingTracks(canvasId, mcolor, ccolor)
            {
                var canvas = document.getElementById(canvasId);
                var context = canvas.getContext('2d');
                
                //reset/clear the canvas
                canvas.width = canvas.width;
                canvas.height = canvas.height;
                 
                //draw main track
                context.beginPath();
                context.moveTo(0,24);
                context.lineTo(48,24);
                context.lineWidth = 6;
                context.strokeStyle = mcolor;
                context.stroke(); 
                
                //draw cross track
                context.beginPath();
                context.moveTo(24,0);
                context.lineTo(24,48);
                context.lineWidth = 6;
                context.strokeStyle = ccolor;
                context.stroke(); 
            }              

            function DrawDiagonalCrossingTracks(canvasId, mcolor, ccolor, rotation)
            {
                var canvas = document.getElementById(canvasId);
                var context = canvas.getContext('2d');
                
                //reset/clear the canvas
                canvas.width = canvas.width;
                canvas.height = canvas.height;
                 
                //draw main track
                context.beginPath();
                context.moveTo(0,0);
                context.lineTo(48,48);
                context.lineWidth = 6;
                context.strokeStyle = mcolor;
                context.stroke(); 
                
                //draw cross track
                context.beginPath();
                context.moveTo(0,48);
                context.lineTo(48,0);
                context.lineWidth = 6;
                context.strokeStyle = ccolor;
                context.stroke(); 
            } 
            
            function DrawCornerTrack(canvasId, color, rotation, flip)
            {               
                var canvas = document.getElementById(canvasId);
                var context = canvas.getContext('2d');
                
                //reset/clear the canvas
                canvas.width = canvas.width;
                canvas.height = canvas.height;
                
                //translate context to center of canvas to support rotation and flipping
                var xoffset = canvas.width/2;
                var yoffset = canvas.height/2;
                context.translate(xoffset, yoffset);
                
                //rotate context
                context.rotate(rotation*Math.PI/180);
                
                //flip horizontal or vertical by changing scale using negative values
                if ("horizontal" == flip)
                {  
                    context.scale(-1, 1);
                }
                
                if ("vertical" == flip)
                {  
                    context.scale(1, -1);
                }

                //draw track
                context.beginPath();
                context.moveTo(0-xoffset,48-yoffset);
                context.lineTo(24-xoffset,24-yoffset);
                context.lineTo(48-xoffset,24-yoffset);
                context.lineJoin = 'miter';
                context.lineWidth = 6;
                context.strokeStyle = color;
                context.stroke();    
            }
            
            function DrawSwitch(canvasId, normal, reverse, tcolor, rotation, flip)
            {                 
                var canvas = document.getElementById(canvasId);
                var context = canvas.getContext('2d');
                
                //reset/clear the canvas
                canvas.width = canvas.width;
                canvas.height = canvas.height;
   
                //translate context to center of canvas to support rotation and flipping
                var xoffset = canvas.width/2;
                var yoffset = canvas.height/2;
                context.translate(xoffset, yoffset);
                
                //rotate context
                context.rotate(rotation*Math.PI/180);
                
                //flip horizontal or vertical by changing scale using negative values
                if ("horizontal" == flip)
                {  
                    context.scale(-1, 1);
                }
                
                if ("vertical" == flip)
                {  
                    context.scale(1, -1);
                }
               
                if (normal)
                {          
                    //draw reverse track
                    context.beginPath();
                    context.moveTo(24-xoffset,24-yoffset);
                    context.lineTo(48-xoffset,0-yoffset);
                    context.lineWidth = 6;
                    context.strokeStyle = defaultTrackColor;
                    context.stroke();    
                
                    //draw normal track
                    context.beginPath();
                    context.moveTo(0-xoffset,24-yoffset);
                    context.lineTo(48-xoffset,24-yoffset);
                    context.lineWidth = 6;
                    context.strokeStyle = tcolor;
                    context.stroke(); 
                
                    //draw line to show normal position
                    context.beginPath();
                    context.moveTo(0-xoffset,20-yoffset);
                    context.lineTo(48-xoffset,20-yoffset);
                    context.lineWidth = 2.5;
                    context.strokeStyle = backgroundColor;
                    context.stroke();   
                }               
                else if (reverse)
                {     
                    //draw normal track
                    context.beginPath();
                    context.moveTo(0-xoffset,24-yoffset);
                    context.lineTo(48-xoffset,24-yoffset);
                    context.lineWidth = 6;
                    context.strokeStyle = defaultTrackColor;
                    context.stroke(); 
                    
                    //draw reverse track starting from switch point              
                    context.beginPath();
                    context.moveTo(0-xoffset,24-yoffset);
                    context.lineTo(24-xoffset,24-yoffset);
                    context.lineTo(48-xoffset,0-yoffset);
                    context.lineJoin = 'miter';
                    context.lineWidth = 6;
                    context.strokeStyle = tcolor;
                    context.stroke(); 
                    
                    //draw line to show reverse position
                    context.beginPath();
                    context.moveTo(26-xoffset,28-yoffset);
                    context.lineTo(34-xoffset,20-yoffset);
                    context.lineWidth = 3;
                    context.strokeStyle = backgroundColor;
                    context.stroke();   
                }
                else
                {
                    //draw reverse track
                    context.beginPath();
                    context.moveTo(24-xoffset,24-yoffset);
                    context.lineTo(48-xoffset,0-yoffset);
                    context.lineWidth = 6;
                    context.strokeStyle = tcolor;
                    context.stroke();    
                
                    //draw normal track
                    context.beginPath();
                    context.moveTo(0-xoffset,24-yoffset);
                    context.lineTo(48-xoffset,24-yoffset);
                    context.lineWidth = 6;
                    context.strokeStyle = tcolor;
                    context.stroke(); 
                }
            }
            
            function DrawSignal(canvasId, scolor, tcolor, rotation, flip)
            {               
                var canvas = document.getElementById(canvasId);
                var context = canvas.getContext('2d');
                
                //reset/clear the canvas
                canvas.width = canvas.width;
                canvas.height = canvas.height;
                
                //translate context to center of canvas to support rotation and flipping
                var xoffset = canvas.width/2;
                var yoffset = canvas.height/2;
                context.translate(xoffset, yoffset);
                
                //rotate context
                context.rotate(rotation*Math.PI/180);
                
                //flip horizontal or vertical by changing scale using negative values
                if ("horizontal" == flip)
                {  
                    context.scale(-1, 1);
                }
                
                if ("vertical" == flip)
                {  
                    context.scale(1, -1);
                }

                //draw track
                context.beginPath();
                context.moveTo(0-xoffset,24-yoffset);
                context.lineTo(48-xoffset,24-yoffset);
                context.lineWidth = 6;
                context.strokeStyle = tcolor;
                context.stroke();    
                
                //draw signal head
                var radius = 6;
                context.beginPath();
                context.arc(21-xoffset, 10-yoffset, radius, 0, 2 * Math.PI, false);
                context.fillStyle = scolor;
                context.fill();
                context.lineWidth = 2;
                context.strokeStyle = scolor;
                context.stroke();
                
                //draw signal mast
                context.beginPath();
                context.moveTo(2-xoffset,10-yoffset);
                context.lineTo(15-xoffset,10-yoffset);
                context.lineWidth = 3;
                context.strokeStyle = defaultTrackColor;
                context.stroke(); 
                
                context.beginPath();
                context.moveTo(2-xoffset,4-yoffset);
                context.lineTo(2-xoffset,16-yoffset);
                context.lineWidth = 3;
                context.strokeStyle = defaultTrackColor;
                context.stroke();
            }
            
            function DrawSignal45(canvasId, scolor, tcolor, rotation, flip)
            {               
                var canvas = document.getElementById(canvasId);
                var context = canvas.getContext('2d');
                
                //reset/clear the canvas
                canvas.width = canvas.width;
                canvas.height = canvas.height;
                
                //translate context to center of canvas to support rotation and flipping
                var xoffset = canvas.width/2;
                var yoffset = canvas.height/2;
                context.translate(xoffset, yoffset);
                
                //rotate context
                context.rotate(rotation*Math.PI/180);
                
                //flip horizontal or vertical by changing scale using negative values
                if ("horizontal" == flip)
                {  
                    context.scale(-1, 1);
                }
                
                if ("vertical" == flip)
                {  
                    context.scale(1, -1);
                }

                //draw track
                context.beginPath();
                context.moveTo(0-xoffset,48-yoffset);
                context.lineTo(48-xoffset,0-yoffset);
                context.lineWidth = 6;
                context.strokeStyle = tcolor;
                context.stroke();    
                
                //draw signal head
                var radius = 6;
                context.beginPath();
                context.arc(21-xoffset, 10-yoffset, radius, 0, 2 * Math.PI, false);
                context.fillStyle = scolor;
                context.fill();
                context.lineWidth = 2;
                context.strokeStyle = scolor;
                context.stroke();
                
                //draw signal mast
                context.beginPath();
                context.moveTo(8-xoffset,23-yoffset);
                context.lineTo(18-xoffset,13-yoffset);
                context.lineWidth = 3;
                context.strokeStyle = defaultTrackColor;
                context.stroke(); 
                
                context.beginPath();
                context.moveTo(2-xoffset,19-yoffset);
                context.lineTo(12-xoffset,29-yoffset);
                context.lineWidth = 3;
                context.strokeStyle = defaultTrackColor;
                context.stroke(); 
            }
            
            function DrawCircle(canvasId, icolor)
            {               
                var canvas = document.getElementById(canvasId);
                var context = canvas.getContext('2d');

                //reset/clear the canvas
                canvas.width = canvas.width;
                canvas.height = canvas.height;
                             
                //draw track
                var radius = 8;
                context.beginPath();
                context.arc(24, 24, radius, 0, 2 * Math.PI, false);
                context.fillStyle = icolor;
                context.fill();
                context.lineWidth = 2;
                context.strokeStyle = icolor;
                context.stroke();
                
            }
            
            function DrawArrow(canvasId, icolor, rotation)
            {               
                var canvas = document.getElementById(canvasId);
                var context = canvas.getContext('2d');

                //reset/clear the canvas
                canvas.width = canvas.width;
                canvas.height = canvas.height;
                
                var xoffset = canvas.width/2;
                var yoffset = canvas.height/2;
                context.translate(xoffset, yoffset);   
                
                //rotate context
                context.rotate(rotation*Math.PI/180);
                
                //draw line
                context.beginPath();
                context.moveTo(10-xoffset,24-yoffset);
                context.lineTo(38-xoffset,24-yoffset);
                context.lineWidth = 6;
                context.strokeStyle = icolor;
                context.stroke();    
                
                //draw point
                context.beginPath();
                context.moveTo(30-xoffset,16-yoffset);
                context.lineTo(38-xoffset,24-yoffset);
                context.lineTo(30-xoffset,32-yoffset);
                context.lineJoin = 'miter';
                context.lineWidth = 6;
                context.strokeStyle = icolor;
                context.stroke();    
            }
            
            function SetLatchedControl(bit, state, btnNbr)
            {              
                //set the control bit
                controlBits[bit] = state;
                
                if (1 == state)
                {                   
                    //disable the button while control is latched
                    document.getElementById("btn_" + btnNbr).disabled = true;
                }
                else
                {
                    //enable the button when control is unlatched
                    document.getElementById("btn_" + btnNbr).disabled = false;
                }
                
                //console.log(Date() + " Controls: " + controlData);              
            }
                        
            function SendControl(bit, state)
            {                 
                QnetClientThread.postMessage({devMode:developerMode, url:hostUrl, sendControl:true, bitNbr:bit, bitState:state, sendAllControls:false, cntrlBits:controlBits});                   
            }
            
            function SendAllControls()
            {
                QnetClientThread.postMessage({devMode:developerMode, url:hostUrl, sendControl:false, bitNbr:0, bitState:0, sendAllControls:true, cntrlBits:controlBits});  
            }
            
            function ClearAllControls()
            {
                //clear all states in the controlBits array
                for (var i=0; i < controlBits.length; i++)
                {
                    controlBits[i] = 0;
                }
                
                //Notify the user that all control bits have been cleared
                $( "#message_dialog" ).dialog( "option", "title", "Clear All Controls" );

                document.getElementById("message_dialog").innerHTML = '<div><img style="float: left; margin-top: 5px; width: 36px" src="Images/info32px.png"/>'
                    + '<div style="float: left; margin-left: 15px; margin-top: 6px; width: 200px">All control bits have been cleared (but not sent).</div></div>';   
                
                ShowMessageDialog();
                
                //Close the controls dialog box if open.
                if ($( "#controls_dialog" ).dialog( "isOpen" ))
                {
                    $( "#controls_dialog" ).dialog( "close" );
                } 
                                
                //console.log(Date() + " Controls: " + controlData);              
            }
            
            function ShowMessageDialog()
            {
                //This function is called to show a message dialog. The innerHTML and title of the dialog
                //are set by the caller
                $( "#message_dialog" ).dialog( "option", "buttons", [
                    { 
                        text: "Close", click: function() { 
                            $( this ).dialog( "close" );
                        }  
                    }
                ]);
                    
                $( "#message_dialog" ).dialog( "open" );
            }
            
            function pad(n)
            {
                return (n < 10) ? '0' + n.toString() : n.toString();
            }

            function SecondsElapsed(startTime)
            {
                var currentTime = new Date();
                var timeDiff = currentTime - startTime; //in ms
                
                // strip the ms
                timeDiff /= 1000;

                // return seconds 
                var seconds = Math.round(timeDiff);
                return seconds;
            }
                                   
            function MillisecondsElapsed(startTime)
            {
                var currentTime = new Date();
                var timeDiff = currentTime - startTime; //in ms
                return timeDiff;
            }
            
            //This function updates the indications for an icon of a mosaic tile design
            function UpdateIconStatus(id, indBits)
            {
                var bit;
                var ind0_state = 0;
                var ind1_state = 0;
                var color1 = defaultTrackColor;
                var color2 = defaultTrackColor;

                //check indication0 status
                if (document.getElementById(id).dataset.indication0.split(",")[0] > 0)
                {
                    //get the bit number and state
                    bit = document.getElementById(id).dataset.indication0.split(",")[0] - 1;
                    //ind0_state = GetBitState(bit);
                    ind0_state = indBits[bit];
                    
                    //invert the state if state option is set to false
                    if (1 == document.getElementById(id).dataset.indication0.split(",")[1])
                    {
                        ind0_state = !ind0_state;
                    }
                }
             
                //check indication1 status
                if (document.getElementById(id).dataset.indication1.split(",")[0] > 0)
                {
                    //get the bit number and state
                    bit = document.getElementById(id).dataset.indication1.split(",")[0] - 1;
                    //ind1_state = GetBitState(bit); 
                    ind1_state = indBits[bit]; 

                    //invert the state if state option is set to false
                    if (1 == document.getElementById(id).dataset.indication1.split(",")[1])
                    {
                        ind1_state = !ind1_state;
                    }
                }
     
                //check the first group of colored indication statuses
                for (var i=2, k=0; i < iconColors.length + 2; i++, k++)
                {
                    if (document.getElementById(id).dataset["indication" + i].split(",")[0] > 0)
                    {
                        //get the bit number and state
                        bit = document.getElementById(id).dataset["indication" + i].split(",")[0] - 1;
                        
                        //var state = GetBitState(bit);
                        var state = indBits[bit];

                        //invert the state if state option is set to false
                        if (1 == document.getElementById(id).dataset["indication" + i].split(",")[1])
                        {
                            state = !state;
                        }                  
                        
                        if (state)
                        {
                            color1 = iconColors[k];
                        }
                    }                    
                }
  
                //check the second group of colored indication statuses
                for (var i=iconColors.length + 2, k=0; i < iconColors.length * 2 + 2; i++, k++ )
                {
                    if (document.getElementById(id).dataset["indication" + i].split(",")[0] > 0)
                    {
                        //get the bit number and state
                        bit = document.getElementById(id).dataset["indication" + i].split(",")[0] - 1;
                        
                        //var state = GetBitState(bit);
                        var state = indBits[bit];

                        //invert the state if state option is set to false
                        if (1 == document.getElementById(id).dataset["indication" + i].split(",")[1])
                        {
                            state = !state;
                        }                  
                        
                        if (state)
                        {
                            color2 = iconColors[k];              
                        }
                    }                
                }
                
                //Now draw the icon with the updated indications
                DrawIcon(id, ind0_state, ind1_state, color1, color2);
            }
            
            //This function is called when the user clicks on the Open file button
            function OpenPanelFile()
            {
                //Send an http GET request to the QLCP-NET Devleoper Tool server so that it will provide the Open File Dialog box.
                //Note: Due to browser security, the developer tool must provide the file path to this control panel.
                //If the the request is successful (i.e. there is a response from the developer tool = "open"),
                //set the requestFileName flag so that this web client will periodically
                //send an http request until the user makes a file selection and the file name is received.
                
                var openrequest = $.get("http://localhost:9999/open/", function(data){
                    if ("open" === data)
                    {
                        requestFileName = true;    
                    }
                }, "text");
                
                //If the request for Open File Dialog fails, show an error message to the user.
                openrequest.fail(function() { 
                    ShowDevToolRequestErrorMessage();  
                });
            }
            
            function OpenPanel(file)
            {  
                var prevIdNbr = 0;
                var tempId = "";
                                
                //Send an http GET to request the panel file contents
                $.get(file, function(data){
                    
                    //We will save the lines of text in the file data in a temporary array. Each line is an array element.
                    
                    //This is what we do for mosaic tile panels (icons and text):
                    //- Look for lines with "icon_begin". This is the start of icon attributes.
                    //- Create a canvas element with the attributes stored as name/value pairs in the array.
                    //- Place the canvas in the div based on the gridlocation attribute value.
                    //- Draw the canvas based on the type as defined by the div value contained in the dataset.idiv attribute    
                    //- As we open the layout file and discover icons, we need to create a new activeIcons array.                   
                    //- After creating the icons, get each text element in the same manner (when we see text_begin).
                    
                    var gridId; //to save grid location when found
                    
                    //Hide the any open dialog box
                    CloseAllDialogBoxes();
                
                    //reset id number
                    idNbr = 0;
                                
                    //clear the activeIcons array
                    activeIcons.length = 0;
                    
                    //clear the component layer
                    component_layer.removeChildren();
                    component_layer.clear();
                    
                    //remove any background image for the panel (component style designs)
                    if (document.getElementById('lcp_background').childElementCount > 0)
                    {
                        document.getElementById('lcp_background').removeChild(document.getElementById('lcp_background').firstChild);
                    }
                
                    //initialize the grid
                    InitGrid();                   

                    //get the lines from the text data and save them in an array
                    var lines = new Array();
                    if (data.indexOf("\r") > 0)
                    {
                        lines = data.split("\r\n");
                    }
                    else if (data.indexOf("\n") > 0)
                    {
                        lines = data.split("\n");
                    }
                    
                    //parse each line of text
                    while (lines.length > 0)
                    {     
                        //get the name/value pair (the attribute on this line)
                        var iconAttr = lines.shift().split(":");
                      
                        //if this is the beginning of icon data, read it until icon_end, then draw the icon
                        if (iconAttr[0] === "icon_begin")
                        {            
                            //start with a template html string for the icon canvas
                            var htmlStr = '<canvas id="icon_template" title="" '
                                + 'data-idiv="" '
                                + 'draggable="true" ondragstart="drag(event)" '
                                + 'ondblclick="" '
                                + 'onmouseup="" '
                                + 'width="" height=""' + customDataStr + '></canvas>';
                
                            //Put the template icon in the temp div so that we can populate the attributes as we
                            //go through the layout file one line at a time.
                            document.getElementById("temp").innerHTML = htmlStr;
                            
                            //get the next line
                            iconAttr = lines.shift().split(":");
                            
                            while (iconAttr[0] !== "icon_end") 
                            {
                                //If this is the id attribute, save it. It cannot be set until we are done since we are getting the
                                //temporary element by its temporary id name.
                                if (iconAttr[0] === "id")
                                {
                                    tempId = iconAttr[1];
                                }
                                //Load the attribute data into the icon template (if this is not a style attribute).
                                //The style attribute is for showing the icon as selected. This is not needed and is
                                //ommitted because additional colons in its attribute value result in an invalid attribute value when split.
                                if ((iconAttr[0] !== "style")&&(iconAttr[0] !== "id"))
                                {
                                    document.getElementById("icon_template").setAttribute(iconAttr[0], iconAttr[1]); 
                                }
                        
                                //save grid location when we find it
                                if (iconAttr[0] === "data-gridlocation")
                                {
                                    //this is to check for a previous bug that save icons without grid locations
                                    if ("" === iconAttr[1])
                                    {
                                        alert("Error in saved panel file. Icon has no grid location.");
                                        return;
                                    }
                                    
                                    gridId = iconAttr[1];
                                }
                                
                                //get the next line
                                iconAttr = lines.shift().split(":");
                            }
                            
                            //now set the id attribute
                            document.getElementById("icon_template").setAttribute("id", tempId); 
                                               
                            //move to this icon's grid location (div id stored as gridlocation attribute)
                            document.getElementById(gridId).innerHTML = document.getElementById("temp").innerHTML;
                        
                            //Draw icon
                            DrawIcon(iconAttr[1], 0, 0, defaultTrackColor, defaultTrackColor); 
                            
                            //add this gridId to the activeIcons array
                            activeIcons.push(iconAttr[1]);
                            
                            //Need to find largest idNbr to be sure we don't duplicate id numbers when new icons
                            //are added to this design.
                            if (Number(iconAttr[1].substring(1)) > idNbr)
                            {
                                idNbr = Number(iconAttr[1].substring(1));
                            }
                                                        
                            //check for duplicate idNbr in case of bug in previous versions
                            if (Number(iconAttr[1].substring(1)) == prevIdNbr)
                            {
                                alert("Duplicate ID Numbers: " + prevIdNbr);
                            }
                            prevIdNbr = Number(iconAttr[1].substring(1));
                        }
                 
                        //if this is the lcp background image (component style designs), load it.
                        if (iconAttr[0] === "lcp_background")
                        { 
                            var img = document.createElement('img'); 
                            img.src = iconAttr[1] + ":" + iconAttr[2];
                            document.getElementById('lcp_background').appendChild(img);
                        }
                        
                        //if this is the lcp background image width, load it.
                        if (iconAttr[0] === "lcp_background_width")
                        { 
                            document.getElementById('lcp_background').firstChild.width = iconAttr[1];
                        }
                        
                        //if this is the beginning of component data, read it and then add the component to the layer
                        //Note: "text_begin" and "text_end" is for older panel files.
                        if ((iconAttr[0] === "component_begin") || (iconAttr[0] === "text_begin"))
                        {   
                            var attributes = new Array();       //This array contains the key/value pairs of attributes read from the file
                            var indicationAttrs = new Array(); 
                            var controlAttrs = new Array(); 

                            //get next line
                            iconAttr = lines.shift().split(":");
                            
                            //while not the end of this component element's attributes
                            while ((iconAttr[0] !== "component_end") && (iconAttr[0] !== "text_end"))
                            {
                                //set the attribute name and value  
                                var attrName = iconAttr[0];
                                var attrValue = iconAttr[1];
                                                
                                //Save the attribute name and value in the attributes array as a key/value pair.
                                attributes[attrName] = attrValue;
                                //console.log(attributes[attrName]);
                                
                                //Save control attributes
                                if (attrName.indexOf('control') >= 0)
                                {
                                    controlAttrs.push(attrValue);
                                }
                                
                                //Save indication attributes
                                if (attrName.indexOf('indication') >= 0)
                                {
                                    indicationAttrs.push(attrValue);
                                }
                                
                                //get next line
                                iconAttr = lines.shift().split(":");
                            }
                            
                            //get the id for this component
                            var thisId = Number(attributes.id.substring(1));
                            
                            //need to find largest idNbr to be sure we don't duplicate id numbers
                            if (Number(attributes.id.substring(1)) > idNbr)
                            {
                                idNbr = Number(attributes.id.substring(1));
                            }

                            //check for duplicate idNbr in case of bug in previous versions
                            if (Number(attributes.id.substring(1)) === prevIdNbr)
                            {
                                alert("Duplicate ID Numbers: " + prevIdNbr);
                            }
                            prevIdNbr = Number(attributes.id.substring(1));
                                 
                            var cmpType;
                            if (typeof attributes.type !== 'undefined')
                                cmpType = attributes.type;
                            else
                                cmpType = "text";
                            
                            //Now, create a new component object with the saved attributes
                            dropMode[cmpType] = true;
                            var Component = AddComponent(Math.floor(Number(attributes.x)), Math.floor(Number(attributes.y)), thisId);
                    
                            //set the button value back to "0"
                            DeselectDropComponentButtons(); 
                                                     
                            SetComponentProperties(Component, attributes);
                            
                            //set the control properties for this component
                            var i;
                            var n = controlAttrs.length;
                            for (var i=0; i < n; i++)
                            {
                                Component.attrs["control" + i] = controlAttrs[i];
                            }   
                
                            //set the indication properties for this component
                            n = indicationAttrs.length;
                            for (i=0; i < n; i++)
                            {
                                Component.attrs["indication" + i] = indicationAttrs[i];
                            }  
                                                                                   
                            //update the selection border (needed for older panels that were saved before selection border was updated)
                            if (typeof Component.find('Text')[0] !== 'undefined') Component.find('Text')[0].setText(Component.find('Text')[0].attrs.text);
                            Component.find('.selection')[0].attrs.update();
                        }
                    }
                    
                    if (designMode)
                    {
                        //if there are icons in this panel, set the panel style to Mosaic.
                        if (activeIcons.length > 0)
                        {
                            document.getElementById("design_style_option").selectedIndex = 1;
                        }
                        else //component style
                        {
                            document.getElementById("design_style_option").selectedIndex = 0;
                        }
                        SetPanelStyle(false);
                    }
                    else //Run Mode
                    {
                        //if there are icons in this panel, move the components container to the background
                        //else, bring it forward using zIndex.
                        if (activeIcons.length > 0)
                        {
                            document.getElementById("component_container").style.zIndex = 3;
                        }
                        else
                        {
                            document.getElementById("component_container").style.zIndex = 10;
                        }
                    }
                    
                    redrawComponents = true;
                    
                }, "text");                 
            }
              
            function SetComponentProperties(Component, attributes)
            {
                var cmpType = Component.attrs.type;
                var indicatorObj;
                var textObj;
                
                if (Component.find('.indicator').length > 0) indicatorObj = Component.find('.indicator')[0];
                if (Component.find('Text').length > 0) textObj = Component.find('Text')[0];
                                        
                //if this is pb_square, set its text properties
                if (cmpType === "pb_square")
                {
                    if (typeof attributes.text !== 'undefined') textObj.setText(attributes.text);
                    if (typeof attributes.fill !== 'undefined') textObj.fill(attributes.fill);
                    if (typeof attributes.fontSize !== 'undefined') textObj.fontSize(attributes.fontSize);
                    textObj.attrs.alignText();  //this has to occur before scaling!
                }

                //if this is text or timer, set its text properties
                if (cmpType === "text" ||
                    cmpType === "timer")
                {
                    if (typeof attributes.text !== 'undefined') textObj.setText(attributes.text);
                    if (typeof attributes.defaultColor !== 'undefined')
                        indicatorObj.attrs.defaultColor = attributes.defaultColor;
                    else
                        indicatorObj.attrs.defaultColor = 'white';  //for older versions that don't support default color.
                    if (typeof attributes.fontSize !== 'undefined') textObj.attrs.fontSize = attributes.fontSize;
                }                           
                else if (typeof attributes.scaleX !== 'undefined')
                {
                    
                    //scale applies to all components but text,
                    //and yes, scaleX is used for y.
                    Component.scale({
                        x: attributes.scaleX,
                        y: attributes.scaleX
                    }); 
            
                }
                
                if (cmpType === "timer")
                {
                    indicatorObj.attrs.timerDuration = attributes.timerDuration;
                }

                //if this is LED, quadrilateral, or arrow, set stroke values.
                if (cmpType === "LED" ||
                    cmpType === "quadrilateral" ||
                    cmpType === "arrow")
                {
                    if (typeof attributes.stroke !== 'undefined') indicatorObj.stroke(attributes.stroke);
                    if (typeof attributes.strokeWidth !== 'undefined') indicatorObj.strokeWidth(attributes.strokeWidth);
                    if (typeof attributes.strokeEnabled !== 'undefined') indicatorObj.strokeEnabled(attributes.strokeEnabled.toString() === 'true');
                    if (typeof attributes.defaultColor !== 'undefined') indicatorObj.attrs.defaultColor = attributes.defaultColor;
                }

                //if this is text, quadrilateral or arrow, set its properties
                if (cmpType === "text" ||
                    cmpType === "quadrilateral" ||
                    cmpType === "arrow" ||
                    cmpType === "timer" ||
                    cmpType === "togSw" ||
                    cmpType === "togSwMom")
                {
                    if (typeof attributes.rotation !== 'undefined') Component.rotation(attributes.rotation);
                }

                //if this is a quadrilateral, set its point coordinates.
                if (cmpType === "quadrilateral")
                {
                    var quad = indicatorObj.attrs;
                    quad.point1x = attributes.point1x;
                    quad.point1y = attributes.point1y;
                    quad.point2x = attributes.point2x;
                    quad.point2y = attributes.point2y;
                    quad.point3x = attributes.point3x;
                    quad.point3y = attributes.point3y;
                    quad.point4x = attributes.point4x;
                    quad.point4y = attributes.point4y;
                }

                if (Component.attrs.nbrCtrls > 0)
                {
                    if (typeof attributes.passcode !== 'undefined') Component.attrs.passcode = attributes.passcode;
                    if (typeof attributes.timeout !== 'undefined') Component.attrs.timeout = attributes.timeout;
                }

                if (Component.attrs.hasIndications)
                {
                    if (typeof attributes.flashTime !== 'undefined') Component.attrs.flashTime = attributes.flashTime;
                }
        
                //component_layer.draw();
            }
            
            //This function is called when the user clicks on the Save button
            function SavePanel()
            {
                saveStr = "version = 3\n";
                
                //save the icons (add the icon data to the save string)
                for (var i=0; i < activeIcons.length; i++)
                {
                    saveStr = saveStr + "icon_begin:" + activeIcons[i] + "\n";
                    try
                    {
                        Array.prototype.slice.call(document.getElementById(activeIcons[i]).attributes).forEach(function(item) {
                            saveStr = saveStr + item.name + ':'+ item.value + "\n";
                        });
                    }
                    catch(e)
                    {
                        alert("Problem with saving icon ID: " + activeIcons[i])
                    }
                    saveStr = saveStr + "icon_end:" + activeIcons[i] + "\n";
                }
                
                //save the text and other components (add their data to the save string)
                var panelObjs = component_layer.getChildren();
                for (var i=0; i < panelObjs.length; i++)
                {
                    saveStr = saveStr + "component_begin:\n";
                    
                    //save all the attributes at the component group level
                    for (key in panelObjs[i].attrs)
                    {
                        saveStr = saveStr + key + ':'+ panelObjs[i].attrs[key] + "\n";
                    } 
                    
                    //save all the attributes for the component's indicator object
                    if (typeof panelObjs[i].find('.indicator')[0] !== 'undefined')
                    {
                        for (key in panelObjs[i].find('.indicator')[0].attrs)
                        {
                            saveStr = saveStr + key + ':'+ panelObjs[i].find('.indicator')[0].attrs[key] + "\n";
                        }     
                    }
                    
                    //save all the attributes for the component's label object
                    if (typeof panelObjs[i].find('.label')[0] !== 'undefined')
                    {
                        for (key in panelObjs[i].find('.label')[0].attrs)
                        {
                            saveStr = saveStr + key + ':'+ panelObjs[i].find('.label')[0].attrs[key] + "\n";
                        }     
                    }
                    
                    saveStr = saveStr + "component_end:\n";
                }           
                                
                //If it exists, save LCP background image as the base64 string that it is.
                if (document.getElementById('lcp_background').childElementCount > 0)
                {
                    saveStr = saveStr + "lcp_background:" + document.getElementById('lcp_background').children[0].src + "\n";
                    saveStr = saveStr + "lcp_background_width:" + document.getElementById('lcp_background').firstChild.width;
                }
            
                //Now, request the QLCP developer tool to open the Save File Dialog box so that the user can select a directory and 
                //enter a file name. Send the data to be saved with the POST request. The rest of the save will be
                //handled by the QLCP developer tool when the user clicks on the Save button of the Save File Dialog box.
                var saverequest = $.post("http://localhost:9999/save/", saveStr, function(){
                }, "text");
                
                //If the request for Save File Dialog fails, show an error message to the user.
                saverequest.fail(function() { 
                    ShowDevToolRequestErrorMessage();  
                });
            }
            
            //This function is called when the developer tool is unavailable (no response) during the Open and Save file functions.
            function ShowDevToolRequestErrorMessage()
            {
                $( "#message_dialog" ).dialog( "option", "title", "Request Timeout" );

                document.getElementById("message_dialog").innerHTML = '<div><img style="float: left; margin-top: 5px; width: 36px" src="Images/error32px.png"/>'
                    + '<div style="float: left; margin-left: 15px; margin-top: 6px; width: 200px">The QLCP-NET Developer Tool is not responding to the request.</div></div>';   
                
                ShowMessageDialog();   
            }
            
            //This function shows a dialog box to the user warning that all icons and text will be cleared.
            function ShowClearAllDialog()
            {
                document.getElementById("clear_all_dialog").innerHTML = '<div><img style="float: left; margin-top: 5px; width: 36px" src="Images/warning_icon.png"/>'
                    + '<div style="float: left; margin-left: 15px; margin-top: 6px; width: 200px">This will delete everything in your current design!</div></div>';

                $( "#clear_all_dialog" ).dialog( "option", "buttons", [
                    { 
                        text: "OK", click: function() { 
                            $( this ).dialog( "close" );
                            ClearAllComponents();
                        }  
                    },
                    { 
                        text: "Cancel", click: function() { 
                            $( this ).dialog( "close" ); 
                        }  
                    }
                ]);
                    
                $( "#clear_all_dialog" ).dialog( "open" );
            }
            
            function ClearAllComponents()
            { 
                idNbr = 0;

                //clear the activeIcons array
                activeIcons.length = 0;

                //clear the components layer
                component_layer.removeChildren();
                component_layer.clear();

                //remove background image for the panel if it exists
                if (document.getElementById('lcp_background').childElementCount > 0)
                {
                    document.getElementById('lcp_background').removeChild(document.getElementById('lcp_background').firstChild);
                }

                //initialize the grid and return to design mode
                InitGrid();                            
                DesignMode();

                if ($( "#properties_dialog" ).dialog( "isOpen" ))
                {
                    $( "#properties_dialog" ).dialog( "close" );
                }                            
            }
           
            //This code supports the drag-n-drop feature for adding an lcp background image.
            if(window.FileReader)
            {
                addEventHandler(window, 'load', function()
                {               
                    var drop = document.getElementById('component_container');

                    function cancel(e) 
                    {
                        if (e.preventDefault) { e.preventDefault(); }
                        return false;
                    }

                    // Tell the browser that we CAN drop on this target
                    addEventHandler(drop, 'dragover', cancel);
                    addEventHandler(drop, 'dragenter', cancel);
                
                    addEventHandler(drop, 'drop', function (e)
                    {
                        //e = e || window.event; // get window.event if e argument missing (in IE)   
                        if (e.preventDefault) { e.preventDefault(); } // stops the browser from redirecting off to the image.

                        var dt = e.dataTransfer;
                        var files = dt.files;
                        
                        for (var i=0; i < files.length; i++)
                        {
                            var file = files[i];
                            var reader = new FileReader();

                            //attach event handlers here...
                            addEventHandler(reader, 'loadend', function(e, file)
                            {
                                //remove any background image for the panel that already exists
                                if (document.getElementById('lcp_background').childElementCount > 0)
                                {
                                    document.getElementById('lcp_background').removeChild(document.getElementById('lcp_background').firstChild);
                                }
                                
                                //add the newly dropped image
                                var bin = this.result; 
                                var img = document.createElement('img'); 
                                img.file = file;   
                                img.src = bin;
                                document.getElementById('lcp_background').appendChild(img);
                                
                            }.bindToEventHandler(file));

                            reader.readAsDataURL(file);
                        }
                    return false;
                    });                                                                   
                });
            }
            else
            { 
                alert("Your browser does not support the HTML5 FileReader.");
            }
            
            //This function is used for dropping the background image
            function addEventHandler(obj, evt, handler)
            {
                if(obj.addEventListener) {
                    // W3C method
                    obj.addEventListener(evt, handler, false);
                } else if(obj.attachEvent) {
                    // IE method.
                    obj.attachEvent('on'+evt, handler);
                } else {
                    // Old school method.
                    obj['on'+evt] = handler;
                }
            }
            
            //This function is used for dropping the background image
            Function.prototype.bindToEventHandler = function bindToEventHandler()
            {
                var handler = this;
                var boundParameters = Array.prototype.slice.call(arguments);
                //create closure
                return function(e)
                {
                    //e = e || window.event; // get window.event if e argument missing (in IE)   
                    boundParameters.unshift(e);
                    handler.apply(this, boundParameters);
                }
            };
           
            function ToggleDropComponentButtons(id)
            {        
                //move the component layer in front of the icons layer using zIndex
                document.getElementById("component_container").style.zIndex = 10;
         
                //Clear any selected icons
                ClearSelection();
                
                //If this is mosaic tile style, close properties dialog box if open
                if(document.getElementById("design_style_option").selectedIndex === 1)
                {
                    if ($( "#properties_dialog" ).dialog( "isOpen" ))
                    {
                        $( "#properties_dialog" ).dialog( "close" );
                    }
                }
                
                //Toggle the drop mode state
                //and show this state by adding or removing the button down class.
                if (dropMode[buttonToComponent[id]])
                {
                    DeselectDropComponentButtons();
                }
                else
                {
                    DeselectDropComponentButtons();
                    $( '#' + id ).addClass("buttonDown");
                    dropMode[buttonToComponent[id]] = true;
                    dropMode["none"] = false;
                }
            }
                        
            function DeselectDropComponentButtons()
            {             
                //clear the values for all of the component buttons
                for (key in buttonToComponent)
                {
                    $( '#' + key ).removeClass("buttonDown");
                }
             
                if (dropMode['icons'] && $( "#icons_dialog" ).dialog( "isOpen" ))
                {
                    $( "#icons_dialog" ).dialog( "close" );
                }
                
                //clear all modes
                for (key in dropMode)
                {
                    dropMode[key] = false;
                }
                
                //set mode to none
                dropMode["none"] = true;
            }
            
//********************* Component functions/classes ***************************************************************************************************************
            
            //All components are groups of shapes that inherit the common properties and methods of the group
            //created in this function.
            function ComponentGroup(xpos, ypos, id_nbr, cmp_type, cmp_text, ctrls, has_indications)
            {
                var Component = new Kinetic.Group({
                    x: xpos,
                    y: ypos,
                    type: cmp_type,
                    text: cmp_text,
                    nbrCtrls: ctrls,
                    hasIndications: has_indications,
                    state: false,
                    scaleX: 1,
                    rotation: 0,
                    flashTime: 500,
                    flashTimeCntr: 0,
                    flashState: 1,
                    passcode: '',
                    unlocked: false,
                    timeout: 0,
                    selected: false,
                    draggable: true,
                    releaseTimeout: 0,
                    id: "c" + id_nbr
                });

                return Component;
            }
            
            //All components inherit the selection border rectangle and sizing handles in this function
            function ComponentSelectionBorder()
            {
                var selectionBorder = new Kinetic.Rect({
                    name: 'selection',
                    fill: 'transparent',
                    stroke: 'cyan',
                    strokeWidth: 2,
                    strokeEnabled: false,

                    update: function()
                    {
                        if (selectionBorder.getParent().attrs.type === 'quadrilateral' ||
                              selectionBorder.getParent().attrs.type === 'LED' ||
                              selectionBorder.getParent().attrs.type === 'text' ||
                              selectionBorder.getParent().attrs.type === 'timer')
                        {
                            var component = selectionBorder.getParent();
                            var borderValues = component.find('.indicator')[0].attrs.getBorderValues();
                            selectionBorder.position({
                                x: borderValues.x,
                                y: borderValues.y
                            });
                            selectionBorder.attrs.width = borderValues.width;
                            selectionBorder.attrs.height = borderValues.height;           
                        }
                    },
                    
                    show: function()
                    {
                        selectionBorder.strokeEnabled(true);
                    },
                    
                    hide: function()
                    {
                        selectionBorder.strokeEnabled(false);
                    }
                });
                
                return selectionBorder;
            }
           
            //This function creates a new quadrilateral.
            function CreateQuadrilateral(xpos, ypos, id_nbr)
            {   
                //create a component that inherits the group component properties
                var Component = ComponentGroup.call(this, xpos, ypos, id_nbr, 'quadrilateral', '', 0, true);
                                
                //create a quadrilateral shape (rectangle initially)
                var quad = new Kinetic.Shape({
                    name: 'indicator', //this is the shape that shows the indications for the component
                    fill: defaultFillColor,
                    defaultColor: defaultFillColor,
                    stroke: 'black',
                    strokeWidth: 0,
                    strokeEnabled: false,
                    point1x: 0,
                    point1y: 0,
                    point2x: 70,
                    point2y: 0,
                    point3x: 70,
                    point3y: 18,
                    point4x: 0,
                    point4y: 18,
                    
                    // a Kinetic.Canvas renderer is passed into the sceneFunc function
                    sceneFunc: function(context)
                    {
                      context.beginPath();
                      context.moveTo(this.attrs.point1x, this.attrs.point1y);
                      context.lineTo(this.attrs.point2x, this.attrs.point2y);
                      context.lineTo(this.attrs.point3x, this.attrs.point3y);
                      context.lineTo(this.attrs.point4x, this.attrs.point4y);
                      context.closePath();
                      context.fillStrokeShape(this);
                    },
                    
                    getBorderValues: function()
                    {
                        var x, y, width, height;
                        
                        x = -quad.attrs.strokeWidth/Component.attrs.scaleX/2;
                        y = -quad.attrs.strokeWidth/Component.attrs.scaleX/2;
                        
                        //for width, use the largest x value
                        if (quad.attrs.point2x > quad.attrs.point3x)
                        {
                            width = Number(quad.attrs.point2x) + (quad.attrs.strokeWidth/Component.attrs.scaleX);
                        }
                        else
                        {
                            width = Number(quad.attrs.point3x) + (quad.attrs.strokeWidth/Component.attrs.scaleX);    
                        }
                        
                        //height, use the largest y value
                        if (quad.attrs.point3y > quad.attrs.point4y)
                        {
                            height = Number(quad.attrs.point3y) + (quad.attrs.strokeWidth/Component.attrs.scaleX);
                        }
                        else
                        {
                            height = Number(quad.attrs.point4y) + (quad.attrs.strokeWidth/Component.attrs.scaleX);
                        }
                        
                        return {
                            x: x,
                            y: y,
                            width: width,
                            height: height                           
                        };   
                    }

                }); 
                
                //we don't want the stroke width to scale
                quad.strokeScaleEnabled(false);
                
                //create a selection border for this component and set its initial width and height
                var selectionBorder = ComponentSelectionBorder.call(this);
                var borderValues = quad.attrs.getBorderValues();
                selectionBorder.position({
                    x: borderValues.x,
                    y: borderValues.y
                });
                selectionBorder.attrs.width = borderValues.width;
                selectionBorder.attrs.height = borderValues.height;
                selectionBorder.strokeScaleEnabled(false);
                                 
                //add shapes to the group
                Component.add(quad, selectionBorder);
                
                //mouse down event
                Component.on('mousedown', function() {
                    ComponentMouseDown(this);
                });
                
                //create a reference for the message listener to be used when we delete this component
                Component.msgListener = function(e){
                    UpdateIndication(e.data.indication, Component, quad);
                };
                
                //add listener for message from web worker
                QnetClientThread.addEventListener("message", Component.msgListener, false);
                                
                return Component;
            }
            
            //This function creates a new arrow.
            function CreateArrow(xpos, ypos, id_nbr)
            { 
                //create a component that inherits the group component properties
                var Component = ComponentGroup.call(this, xpos, ypos, id_nbr, 'arrow', '', 0, true);
                
                //create an arrow
                var arrow = new Kinetic.Shape({
                    name: 'indicator',
                    fill: defaultFillColor,
                    defaultColor: defaultFillColor,
                    stroke: 'black',
                    strokeWidth: 0,
                    strokeEnabled: false,
                    
                    // a Kinetic.Canvas renderer is passed into the sceneFunc function
                    sceneFunc: function(context)
                    {
                      context.beginPath();
                      context.moveTo(0,0);
                      context.lineTo(20,0);
                      context.lineTo(20,-5);
                      context.lineTo(40,5);
                      context.lineTo(20,15);
                      context.lineTo(20,10);
                      context.lineTo(0,10);
                      context.closePath();
                      context.fillStrokeShape(this);
                    }
                }); 
                
                arrow.strokeScaleEnabled(false);
                     
                //create a highlight for this component that inherits the common component highlight properties
                var selectionBorder = ComponentSelectionBorder.call(this);         
                selectionBorder.attrs.y = -5;
                selectionBorder.attrs.width = 40;
                selectionBorder.attrs.height = 20;
                selectionBorder.strokeScaleEnabled(false);
                                 
                //add shapes to the group
                Component.add(arrow, selectionBorder);
                
                //mouse down event
                Component.on('mousedown', function() {
                    ComponentMouseDown(this);
                });
                
                //create a reference for the message listener to be used when we delete this component
                Component.msgListener = function(e){
                    UpdateIndication(e.data.indication, Component, arrow);
                };
                
                //add listener for message from web worker
                QnetClientThread.addEventListener("message", Component.msgListener, false);            
                
                return Component;
            }
            
            //This function creates a new LED.
            //Attributes and functions are created, including the behavior of the selection border and sizing handles.
            function CreateLED(xpos, ypos, id_nbr)
            {    
                //create a component that inherits the group component properties
                var Component = ComponentGroup.call(this, xpos, ypos, id_nbr, 'LED', '', 0, true);
                
                //create the led shape
                var led = new Kinetic.Circle({
                    name: 'indicator',
                    radius: 12,
                    fill: defaultFillColor,
                    defaultColor: defaultFillColor,
                    stroke: 'black',
                    strokeWidth: 2,
                    strokeEnabled: true,  
                                        
                    getBorderValues: function()
                    {
                        var x, y, width, height;
                        
                        x = -(Number(led.attrs.radius) + led.attrs.strokeWidth/Component.attrs.scaleX/2);
                        y = -(Number(led.attrs.radius) + led.attrs.strokeWidth/Component.attrs.scaleX/2);
                        width = 2 * Number(led.attrs.radius) + led.attrs.strokeWidth/Component.attrs.scaleX;
                        height = width;  
                        
                        return {
                            x: x,
                            y: y,
                            width: width,
                            height: height                           
                        };   
                    }

                });
                
                led.strokeScaleEnabled(false);
                //led.perfectDrawEnabled(false); //this could help performance
                
                //create a selection border for this component and set its initial position, width, and height
                var selectionBorder = ComponentSelectionBorder.call(this);
                var borderValues = led.attrs.getBorderValues();
                selectionBorder.position({
                    x: borderValues.x,
                    y: borderValues.y
                });
                selectionBorder.attrs.width = borderValues.width;
                selectionBorder.attrs.height = borderValues.height;
                selectionBorder.strokeScaleEnabled(false);

                //add shapes to the group
                Component.add(led, selectionBorder);
                
                //mouse down event
                Component.on('mousedown', function() {
                    ComponentMouseDown(this);
                });
                
                //create a reference for the message listener to be used when we delete this component
                Component.msgListener = function(e){
                    UpdateIndication(e.data.indication, Component, led);
                };
                
                //add listener for message from web worker
                QnetClientThread.addEventListener("message", Component.msgListener, false);
                                
                return Component;
            }
                       
            //This function creates a new pushbutton.
            function CreatePushbutton(xpos, ypos, id_nbr)
            {  
                //create a component that inherits the group component properties
                var Component = ComponentGroup.call(this, xpos, ypos, id_nbr, 'pb', '', 2, true);

                var swBaseRing = new Kinetic.Circle({
                    radius: 15,
                    fill: 'lightgray',
                    stroke: 'black',
                    strokeWidth: 1
                 });

                var swButton = new Kinetic.Circle({
                    name: 'indicator',
                    radius: 9,
                    fill: defaultFillColor,
                    defaultColor: defaultFillColor,
                    stroke: 'black',
                    strokeWidth: 2
                 });
                                                   
                //create a selection border for this component and initialize its values
                var selectionBorder = ComponentSelectionBorder.call(this);         
                selectionBorder.attrs.x = -swBaseRing.attrs.radius;
                selectionBorder.attrs.y = -swBaseRing.attrs.radius;
                selectionBorder.attrs.width = swBaseRing.attrs.radius * 2;
                selectionBorder.attrs.height = swBaseRing.attrs.radius * 2;
                selectionBorder.strokeScaleEnabled(false);

                Component.add(swBaseRing, swButton, selectionBorder);
                
                //mouse down event
                Component.on('mousedown', function() {
                    ComponentMouseDown(this);
                    if (designMode === false)
                    {
                        CheckLockStatus(this);
                        lastClickTime = new Date();  
                    
                        if (this.attrs.unlocked)                  
                        {
                            swButton.attrs.radius = 8;
                            swButton.attrs.strokeWidth = 3;
                            redrawComponents = true;
                            SendControl(this.attrs["control0"]-1, 1);
                            //console.log('mousedown');
                        }
                    }
                });

                //mouse up event
                Component.on('mouseup', function() {
                    if (designMode === false && this.attrs.unlocked)
                    {
                        swButton.attrs.radius = 9;
                        swButton.attrs.strokeWidth = 2;
                        redrawComponents = true;  
                        SendControl(this.attrs["control0"]-1, 0);
                        //console.log('mouseup');
                    }
                });

                //mouse leave event
                Component.on('mouseleave', function() {
                    if (designMode === false && this.attrs.unlocked)
                    {
                        swButton.attrs.radius = 9;
                        swButton.attrs.strokeWidth = 2;
                        redrawComponents = true;
                        SendControl(this.attrs["control0"]-1, 0);
                        //console.log('mouseleave');
                    }
                });
                
                //create a reference for the message listener to be used when we delete this component
                Component.msgListener = function(e){
                    UpdateIndication(e.data.indication, Component, swButton);
                };
                
                //add listener for message from web worker
                QnetClientThread.addEventListener("message", Component.msgListener, false);
                
                return Component;
            }
            
            //This function creates a new square pushbutton.
            function CreateSquarePushbutton(xpos, ypos, id_nbr)
            {  
                //create a component that inherits the group component properties
                var Component = ComponentGroup.call(this, xpos, ypos, id_nbr, 'pb_square', '', 2, true);

                var swBase = new Kinetic.Rect({
                    width: 65,
                    height: 30,
                    fill: defaultFillColor,
                    stroke: 'black',
                    strokeWidth: 1
                });

                var swButton = new Kinetic.Rect({
                    name: 'indicator',
                    offsetX: -5,
                    offsetY: -5,
                    width: 55,
                    height: 20,
                    //fill: defaultFillColor,
                    defaultColor: 'black',
                    stroke: 'lightgray',
                    strokeWidth: 2
                });
                
                var swText = new Kinetic.Text({
                    name: 'label',
                    text: '',
                    fontSize: 10,
                    fontFamily: 'Arial',
                    fill: 'white',

                    alignText: function()
                    {
                        swText.offsetX(-swBase.getWidth()/2 + swText.getWidth()/2);
                        swText.offsetY(-swBase.getHeight()/2 + swText.getHeight()/2);
                    }
                });
                                               
                //create a selection border for this component and initialize its values
                var selectionBorder = ComponentSelectionBorder.call(this);         
                selectionBorder.attrs.width = swBase.getWidth();
                selectionBorder.attrs.height = swBase.getHeight();
                selectionBorder.strokeScaleEnabled(false);
                
                Component.add(swBase, swButton, swText, selectionBorder);
                
                //Add button pressed event listeners.
                //If it is a touch event, we don't want the mouse event to be processed also so stop propagation/bubbling.
                //(touch events occur first)
                Component.addEventListener('mousedown', function() {btnPressedHandler(this);});
                Component.addEventListener('touchstart', function(e) {btnPressedHandler(this); e.stopPropagation();});
                
                var btnPressedHandler = function (cmp) {
                    ComponentMouseDown(cmp);
                    if (designMode === false)
                    {
                        CheckLockStatus(cmp);
                        lastClickTime = new Date();  
                        
                        if (cmp.attrs.unlocked)
                        {
                            swButton.attrs.width = 53;
                            swButton.attrs.height = 18;
                            swButton.attrs.strokeWidth = 3;
                            swButton.offsetX(-6);
                            swButton.offsetY(-6);
                            swBase.attrs.fill = 'red';
                            redrawComponents = true;
                            SendControl(cmp.attrs["control0"]-1, 1);
                        }
                        
                        //Set a timeout for this button if the user does not release or moves/slides off the button.
                        //This is needed for touch events.
                        cmp.releaseTimeout = setTimeout(function(){btnReleasedHandler(cmp);}, 5000);
                    }
                };
 
                //Add button released event listeners.
                //If it is a touch event, we don't want the mouse event to be processed so stop propagation/bubbling.                
                //(touch events occur first)
                Component.addEventListener('mouseup', function() {btnReleasedHandler(this);});
                Component.addEventListener('mouseleave', function() {btnReleasedHandler(this);});
                Component.addEventListener('touchend', function(e) {btnReleasedHandler(this); e.stopPropagation();});
                
                var btnReleasedHandler = function (cmp) {
                    if (designMode === false && cmp.attrs.unlocked)
                    {
                        swButton.attrs.width = 55;
                        swButton.attrs.height = 20;
                        swButton.attrs.strokeWidth = 2;
                        swButton.offsetX(-5);
                        swButton.offsetY(-5);
                        swBase.attrs.fill = defaultFillColor;
                        redrawComponents = true;
                        SendControl(cmp.attrs["control0"]-1, 0);
                        
                        //cancel the release timeout
                        clearTimeout(cmp.releaseTimeout);
                    }
                };
                                
                //create a reference for the message listener to be used when we delete this component
                Component.msgListener = function(e){
                    UpdateIndication(e.data.indication, Component, swButton);
                };
                
                //add listener for message from web worker
                QnetClientThread.addEventListener("message", Component.msgListener, false);
                
                return Component;
            }
            
            //This function creates a new toggle switch.
            function CreateToggleSwitch(xpos, ypos, id_nbr)
            {  
                //create a component that inherits the group component properties
                var Component = ComponentGroup.call(this, xpos, ypos, id_nbr, 'togSw', '', 3, false);
                
                var swBaseRing = new Kinetic.Circle({   
                    radius: 17,
                    fill: 'lightgray',
                    stroke: 'black',
                    strokeWidth: 1
                 });
                 
                //swBaseRing.perfectDrawEnabled(false);  //this could help performance

                var swBaseCntr = new Kinetic.Circle({    
                    radius: 7,
                    fill: 'black',
                    stroke: 'black',
                    strokeWidth: 1
                 });

                //swBaseCntr.perfectDrawEnabled(false); //this could help performance

                var switchUp = new Kinetic.Shape({
                    name: 'up',
                    fill: 'lightgray',
                    stroke: 'black',
                    visible: false,
                    strokeWidth: 1,         
                    // a Kinetic.Canvas renderer is passed into the sceneFunc function
                    sceneFunc: function(context)
                    {
                      context.beginPath();
                      context.moveTo(-5, 0);
                      context.lineTo(-12, -25);
                      context.lineTo(12, -25);
                      context.lineTo(5, 0);
                      context.bezierCurveTo(5,7,-5,7,-5,0);
                      context.moveTo(-12, -25);
                      context.lineTo(-12, -30);
                      context.lineTo(12,-30);
                      context.lineTo(12,-25);
                      context.fillStrokeShape(this);
                    }
                });
                
                //switchUp.perfectDrawEnabled(false); //this could help performance

                var switchDown = switchUp.clone(); 
                switchDown.name('down');
                switchDown.rotate(180);
                //switchDown.perfectDrawEnabled(false); //this could help performance

                var switchCenter = new Kinetic.Shape({
                    name: 'center',
                    fill: 'lightgray',
                    stroke: 'black',
                    visible: false,
                    strokeWidth: 1,         
                    // a Kinetic.Canvas renderer is passed into the sceneFunc function
                    sceneFunc: function(context)
                    {
                      context.beginPath();
                      context.moveTo(-12, 5);
                      context.lineTo(-12, -5);
                      context.lineTo(12, -5);
                      context.lineTo(12, 5);
                      context.lineTo(-12, 5);
                      context.fillStrokeShape(this);
                    }
                });
                
                //switchCenter.perfectDrawEnabled(false); //this could help performance

                //create a selection border for this component and initialize its values
                var selectionBorder = ComponentSelectionBorder.call(this);         
                selectionBorder.attrs.x = -swBaseRing.attrs.radius;
                selectionBorder.attrs.y = -swBaseRing.attrs.radius;
                selectionBorder.attrs.width = swBaseRing.attrs.radius * 2;
                selectionBorder.attrs.height = swBaseRing.attrs.radius * 2;
                selectionBorder.strokeScaleEnabled(false);

                // Add the component pieces to the group.
                Component.add(swBaseRing, swBaseCntr, switchUp, switchDown, switchCenter, selectionBorder);

                // When adding component from saved file...
                // If this is a momentary toggle switch, change the type
                if (dropMode.togSwMom)
                {
                    Component.attrs.type = "togSwMom";
                    switchCenter.visible(true);
                }
                else
                {
                    switchDown.visible(true);
                }

                //mouse down event
                Component.on('mousedown', function(e) {
                    ComponentMouseDown(this);
                    if (designMode === false || initLatchedSwitches === true)
                    {
                        //e.cancelBubble = true; //no need to bubble this event
                        CheckLockStatus(this);
                        lastClickTime = new Date();  
                        if (this.attrs.unlocked && this.attrs.type === "togSw")
                        {                               
                            if (this.attrs.state)
                            {
                                //turn the switch off
                                this.attrs.state = false;
                                switchUp.visible(false);
                                switchDown.visible(true);
                                controlBits[this.attrs["control0"]-1] = 0;
                                controlBits[this.attrs["control1"]-1] = 1;
                                if (initLatchedSwitches === false) SendAllControls();
                            }
                            else
                            {
                                this.attrs.state = true;
                                switchUp.visible(true);
                                switchDown.visible(false);
                                controlBits[this.attrs["control0"]-1] = 1;
                                controlBits[this.attrs["control1"]-1] = 0;
                                if (initLatchedSwitches === false) SendAllControls();
                            }
                            redrawComponents = true;
                        }
                        if (this.attrs.unlocked && this.attrs.type === "togSwMom")
                        {
                            //If not rotated and clicked on upper side of switch,
                            //Or if rotated and clicked on right side of switch, send control0.
                            if (this.attrs.rotation == 0 && (stage.getPointerPosition().y < this.y())
                                || this.attrs.rotation == 90 && (stage.getPointerPosition().x > this.x()))
                            {
                                switchUp.visible(true);
                                switchDown.visible(false);
                                switchCenter.visible(false);
                                SendControl(this.attrs["control0"]-1, 1);
                            }
                            else //If clicked on lower side of switch, send control1
                            {
                                switchUp.visible(false);
                                switchDown.visible(true);
                                switchCenter.visible(false);
                                SendControl(this.attrs["control1"]-1, 1);                                    
                            } 
                            redrawComponents = true;
                        }                       
                    }
                });
                
                //mouse up event
                Component.on('mouseup', function() {
                    if (designMode === false && this.attrs.unlocked && this.attrs.type === "togSwMom")
                    {
                        switchUp.visible(false);
                        switchDown.visible(false);
                        switchCenter.visible(true);
                        redrawComponents = true;
                        SendControl(this.attrs["control0"]-1, 0);     
                        SendControl(this.attrs["control1"]-1, 0);  
                    }
                });                
               
                //mouse leave event
                Component.on('mouseleave', function() {
                    if (designMode === false && this.attrs.unlocked && this.attrs.type === "togSwMom")
                    {                               
                        switchUp.visible(false);
                        switchDown.visible(false);
                        switchCenter.visible(true);
                        redrawComponents = true;
                        SendControl(this.attrs["control0"]-1, 0);     
                        SendControl(this.attrs["control1"]-1, 0);     
                    }
                });
                
                return Component;
            } 
                                       
            //This function creates a new timer component.
            function CreateTimer(xpos, ypos, id_nbr)
            { 
                //create a component that inherits the group component properties
                var Component = ComponentGroup.call(this, xpos, ypos, id_nbr, 'timer', '', 0, true);

                //create a new text object
                var timer = new Kinetic.Text({
                    name: 'indicator',
                    text: "0s",
                    timerDuration: 0,
                    startTime: 0,
                    timerRunning: false,
                    timerComplete: false,
                    fontSize: 20,
                    fontFamily: 'Arial',
                    fill: 'white',
                    defaultColor: 'white',
                                                                                    
                    getBorderValues: function()
                    {
                        var x, y, width, height;
                        
                        x = 0;
                        y = 0;
                        width = timer.getWidth();
                        height = timer.getHeight();  
                        
                        return {
                            x: x,
                            y: y,
                            width: width,
                            height: height                           
                        };   
                    }
                });
                
                //create a selection border for this component and initialize its values
                var selectionBorder = ComponentSelectionBorder.call(this);         
                selectionBorder.attrs.width = timer.getWidth();
                selectionBorder.attrs.height = timer.getHeight();
                selectionBorder.strokeScaleEnabled(false);
                
                // Add the component pieces to the group.
                Component.add(timer, selectionBorder);
                
                //mouse down event
                Component.on('mousedown', function() {
                    ComponentMouseDown(this);
                });
                
                //create a reference for the message listener to be used when we delete this component
                Component.msgListener = function(e){
                    UpdateIndication(e.data.indication, Component, timer);
                };
                
                //add listener for message from web worker
                QnetClientThread.addEventListener("message", Component.msgListener, false);
                
                return Component;
            }
            
            //This function creates a new text component.
            function CreateText(xpos, ypos, id_nbr)
            { 
                //create a component that inherits the group component properties
                var Component = ComponentGroup.call(this, xpos, ypos, id_nbr, 'text', '', 0, true);

                //create a new text object
                var cmpText = new Kinetic.Text({
                    name: 'indicator',
                    text: 'text',
                    fontSize: 14,
                    fontFamily: 'Arial',
                    fill: 'red',
                    defaultColor: 'red',
                                                            
                    getBorderValues: function()
                    {
                        var x, y, width, height;
                        
                        x = 0;
                        y = 0;
                        width = cmpText.getWidth();
                        height = cmpText.getHeight();  
                        
                        return {
                            x: x,
                            y: y,
                            width: width,
                            height: height                           
                        };   
                    }
                });
                
                //create a selection border for this component and initialize its values
                var selectionBorder = ComponentSelectionBorder.call(this);         
                selectionBorder.attrs.width = cmpText.getWidth();
                selectionBorder.attrs.height = cmpText.getHeight();
                selectionBorder.strokeScaleEnabled(false);

                // Add the component pieces to the group.
                Component.add(cmpText, selectionBorder);
                
                //mouse down event
                Component.on('mousedown', function() {
                    ComponentMouseDown(this);
                });
                
                //create a reference for the message listener to be used when we delete this component
                Component.msgListener = function(e){
                    UpdateIndication(e.data.indication, Component, cmpText);
                };
                
                //add listener for message from web worker
                QnetClientThread.addEventListener("message", Component.msgListener, false);
                
                return Component;
            }
            
            //This function creates a new keyswitch.
            function CreateKeySwitch(xpos, ypos, id_nbr)
            {  
                //create a component that inherits the group component properties
                var Component = ComponentGroup.call(this, xpos, ypos, id_nbr, 'keyswitch', '', 4, false);

                var swBaseRing = new Kinetic.Circle({   
                    radius: 25,
                    fill: 'lightgray',
                    stroke: 'black',
                    strokeWidth: 1
                 });

                var swBaseInnerRing = new Kinetic.Circle({   
                    radius: 20,
                    fill: 'lightgray',
                    stroke: 'black',
                    strokeWidth: 1
                 });
                 
                //up and initial position
                var position1 = new Kinetic.Rect({
                    name: 'pos1',
                    fill: 'black',
                    visible: true,
                    strokeWidth: 0, 
                    x: -4,
                    y: -20,
                    width: 8,
                    height: 25
                });  
                
                //right position
                var position2 = new Kinetic.Rect({
                    name: 'pos2',
                    fill: 'black',
                    visible: false,
                    strokeWidth: 0,   
                    x: -5,
                    y: -4,
                    width: 25,
                    height: 8
                });
                                                
                //left position
                var position3 = new Kinetic.Rect({
                    name: 'pos3',
                    fill: 'black',
                    visible: false,
                    strokeWidth: 0, 
                    x: -20,
                    y: -4,
                    width: 25,
                    height: 8,
                });
                   
                //create a selection border for this component and initialize its values
                var selectionBorder = ComponentSelectionBorder.call(this);         
                selectionBorder.attrs.x = -swBaseRing.attrs.radius;
                selectionBorder.attrs.y = -swBaseRing.attrs.radius;
                selectionBorder.attrs.width = swBaseRing.attrs.radius * 2;
                selectionBorder.attrs.height = swBaseRing.attrs.radius * 2;
                selectionBorder.strokeScaleEnabled(false);

                // Add the component pieces to the group.
                Component.add(swBaseRing, swBaseInnerRing, position1, position2, position3, selectionBorder);
                
                //mouse down event
                Component.on('mousedown', function() {
                    ComponentMouseDown(this);
                    if (designMode === false || initLatchedSwitches === true)
                    {
                        CheckLockStatus(this);
                        lastClickTime = new Date();  
                        
                        if (this.attrs.unlocked)
                        {   
                            if (initLatchedSwitches)
                            {
                                //turn switch to position 1 (up)
                                this.attrs.state = false;
                                position1.visible(true);
                                position2.visible(false);
                                position3.visible(false);
                                controlBits[this.attrs["control0"]-1] = 1;
                                controlBits[this.attrs["control1"]-1] = 0;
                                controlBits[this.attrs["control2"]-1] = 0;
                            }
                            //If clicked on upper side of switch or initializing switch position...
                            else if (stage.getPointerPosition().y < this.y() - 5)
                            {
                                //turn switch to position 1 (up)
                                this.attrs.state = false;
                                position1.visible(true);
                                position2.visible(false);
                                position3.visible(false);
                                controlBits[this.attrs["control0"]-1] = 1;
                                controlBits[this.attrs["control1"]-1] = 0;
                                controlBits[this.attrs["control2"]-1] = 0;
                                SendAllControls();
                            }
                            //if the click was on the right side of the switch
                            else if (stage.getPointerPosition().x > this.x())
                            {
                                //turn switch to position 2 (right)
                                this.attrs.state = true;
                                position1.visible(false);
                                position2.visible(true);
                                position3.visible(false);
                                controlBits[this.attrs["control0"]-1] = 0;
                                controlBits[this.attrs["control1"]-1] = 1;
                                controlBits[this.attrs["control2"]-1] = 0;
                                SendAllControls();
                            }
                            //if the click was on the left and the bit for this control has been assigned...
                            else if (stage.getPointerPosition().x < this.x() && this.attrs["control2"] > 0)
                            {
                                //turn switch to position 3 (left)
                                this.attrs.state = true;
                                position1.visible(false);
                                position2.visible(false);
                                position3.visible(true);
                                controlBits[this.attrs["control0"]-1] = 0;
                                controlBits[this.attrs["control1"]-1] = 0;
                                controlBits[this.attrs["control2"]-1] = 1;
                                SendAllControls();
                                console.log(this.attrs["control0"]);
                                console.log(this.attrs["control1"]);
                                console.log(this.attrs["control2"]);
                            }
                            
                            redrawComponents = true;
                        }
                    }
                });
                
                return Component;
            } 
            
            function ComponentMouseDown(component)
            {
                //if in developer mode...
                if (developerMode)
                {
                    if (designMode)
                    {
                        ShowSelectedComponent(component); 
                        component.draggable(true); 
                        mouseDownOnComponent = true;
                    }
                    else
                    {
                        component.draggable(false);
                    }
                }
            }
            
            //This function updates the compnonent indications and sets the redrawComponents flag if any indication has changed
            function UpdateIndication(indicationBits, componentObj, indicator)
            {
                var color1;
                var bit;
                var state = 0;
                var timerComponent;
                var timerState = 0;
                 
                color1 = indicator.attrs.defaultColor;

                //check the first group of colored indication statuses
                if (designMode === false)
                {  
                    //Update the flash counter and...
                    componentObj.attrs.flashTimeCntr++;
                        
                    //invert the state if flash time has elapsed. Timers are 100 ms resolution
                    if (componentObj.attrs.flashTimeCntr * 100 >= componentObj.attrs.flashTime)
                    {
                        componentObj.attrs.flashTimeCntr = 0;
                        componentObj.attrs.flashState = !componentObj.attrs.flashState;
                    }

                    //set color based on the indication bits
                    var indNbr = 0;
                    for (var i=0; i < iconColors.length; i++)
                    {
                        if (componentObj.attrs["indication" + indNbr].split(",")[0] > 0)
                        {
                            //get the bit number and state
                            bit = componentObj.attrs["indication" + indNbr].split(",")[0] - 1;

                            state = indicationBits[bit];

                            //invert the state if state option is set to false
                            if (1 === Number(componentObj.attrs["indication" + indNbr].split(",")[1]))
                            {
                                state = !state;
                            }  

                            //set the timer state
                            if (componentObj.attrs.type === "timer")
                            {
                                timerState = timerState | state;
                            }

                            //set the fill color based on the state value
                            if (state)
                            {
                                color1 = iconColors[i];
                            }
                        }
                        indNbr++;
                    }

                    //set color based on the flashing indication bits
                    for (var i=0; i < iconColors.length; i++)
                    {
                        if (componentObj.attrs["indication" + indNbr].split(",")[0] > 0)
                        {
                            //get the bit number and state
                            bit = componentObj.attrs["indication" + indNbr].split(",")[0] - 1;

                            state = indicationBits[bit];

                            //invert the state if state option is set to false
                            if (1 === Number(componentObj.attrs["indication" + indNbr].split(",")[1]))
                            {
                                state = !state;
                            } 

                            //set the timer state
                            if (componentObj.attrs.type === "timer")
                            {
                                timerState = timerState | state;
                            }

                            state &= componentObj.attrs.flashState;
                            //console.log("state = " + state);
                            
                            if (state)
                            {
                                color1 = iconColors[i];  
                            }
                        }                    
                        indNbr++;
                    }

                    //if this is a timer component, set the timer values based on the state value
                    if (componentObj.attrs.type === "timer")
                    {
                        timerComponent = indicator.attrs;

                        if (timerState === 1)
                        {
                            if (timerComponent.timerRunning === false && timerComponent.timerComplete === false)
                            {
                                timerComponent.timerRunning = true;
                                timerComponent.startTime = new Date();
                            }
                        }
                        else
                        {
                            //This means that there are no indication bits associated with this timer that are true.
                            timerComponent.timerRunning = false;
                            timerComponent.timerComplete = false;
                        }
                    }  

                    //Update the timer component's displayed time value
                    if (componentObj.attrs.type === "timer")
                    {
                        timerComponent = indicator.attrs;

                        if(timerComponent.timerRunning)
                        {
                            var elapsedTime = SecondsElapsed(timerComponent.startTime); 

                            //Check if timer has elapsed, else continue to decrement each second.
                            if (elapsedTime >= timerComponent.timerDuration)
                            {
                                indicator.setText('0s');
                                timerComponent.timerComplete = true;
                            }
                            else //decrement displayed value
                            {
                                indicator.setText((timerComponent.timerDuration - elapsedTime).toString() + 's');
                            }

                            redrawComponents = true;
                        }
                        else
                        {
                            indicator.setText(timerComponent.timerDuration.toString() + 's');
                        }                         

                    } 

                }

                //If the color changed, update the color and redraw the component.
                if (indicator.attrs.fill !== color1)
                {
                    indicator.attrs.fill = color1;
                    redrawComponents = true;
                }

            }
            
//***********************************************************************************************************************************************************************            
            
            //Components are LEDs, Pushbuttons, toggle switches, text, etc. that can be placed anywhere on the page
            //to create a virtual panel that simulates the traditional hardware-based CLCPs.
            function AddComponent(xpos, ypos, id_nbr)
            {    
                var Component;
                var configureComponent;
                               
                if (dropMode.quadrilateral)
                {
                    Component = CreateQuadrilateral(xpos, ypos, id_nbr);
                    Component.draggable(false);
                    configureComponent = true;
                    if (!developerMode) Component.listening(false);
                }
                
                //create an arrow shape
                if (dropMode.arrow)
                {
                    Component = CreateArrow(xpos, ypos, id_nbr);
                    Component.draggable(false);
                    configureComponent = true;
                    if (!developerMode) Component.listening(false);
                }
                
                //create an LED
                if (dropMode.LED)
                {
                    Component = CreateLED(xpos, ypos, id_nbr);
                    Component.draggable(false);
                    configureComponent = true;
                    if (!developerMode) Component.listening(false);
                }

                //creat a round pushbutton
                if (dropMode.pb)
                {
                    Component = CreatePushbutton(xpos, ypos, id_nbr);
                    Component.draggable(false);
                    configureComponent = true;
                }
                
                //create a square pushbutton
                if (dropMode.pb_square)
                {  
                    Component = CreateSquarePushbutton(xpos, ypos, id_nbr);
                    Component.draggable(false);
                    configureComponent = true;
                }
                
                //create a toggle switch
                if (dropMode.togSw || dropMode.togSwMom)
                {
                    Component = CreateToggleSwitch(xpos, ypos, id_nbr);
                    Component.draggable(false);
                    configureComponent = true;
                }
                
                //create a keyswitch
                if (dropMode.keyswitch)
                {
                    Component = CreateKeySwitch(xpos, ypos, id_nbr);
                    Component.draggable(false);
                    configureComponent = true;
                }
                
                //create text
                if (dropMode.text)
                {  
                    Component = CreateText(xpos, ypos, id_nbr);
                    Component.draggable(false);
                    configureComponent = true;
                    if (!developerMode) Component.listening(false);
                }
                
                //create timer text
                if (dropMode.timer)
                {
                    Component = CreateTimer(xpos, ypos, id_nbr);
                    Component.draggable(false);
                    configureComponent = true;
                    if (!developerMode) Component.listening(false);
                }
                
                //If a new component was added, configure its properties and events
                if (configureComponent)
                {
                    //initialize the indication properties for this component
                    //An indication bit can be assigned for each color.
                    for (var i=0; i < iconColors.length * 2; i++)
                    {
                        Component.attrs["indication" + i] = "0";
                    }

                    //initialize the control properties for this component
                    for (var i=0; i < Component.attrs.nbrCtrls; i++)
                    {
                        Component.attrs["control" + i] = "0";
                    }

                    //Add the dragend event to this component so that it can be dragged to the trash at a later time
                    Component.on('dragend', function() {
                        /*
                        var trashCan_x = 46;
                        var trashCan_y = 46;

                        //If the component location is over the trash can, remove it from the layer.
                        if (this.attrs.x + this.getWidth()/2 < trashCan_x & this.attrs.y < trashCan_y)
                        {
                            if ($( "#properties_dialog" ).dialog( "isOpen" ))
                            {
                                $( "#properties_dialog" ).dialog( "close" );
                            }
                            ClearPreviousSelection();
                            this.destroy(); 
                            component_layer.draw();                           
                            //ClearPositionDisplay();
                        }

                        if ($( "#properties_dialog" ).dialog( "isOpen" ))
                        {
                            ComponentProperties(this);
                        }
                        
                        //console.log('Event: dragend, Selected Id: ' + selectedId);
                        */
                    });    

                    //add the component to the layer
                    component_layer.add(Component);
                    //component_layer.draw();
                }
                
                //return component object
                return Component;
                             
            } //add component
            
            function CheckLockStatus(component)
            {                     
                //Check if passcode is defined and/or assigned for this control component
                //and then determine if it is locked.
                if (typeof component.attrs.passcode === 'undefined'
                        || component.attrs.passcode === null
                        || component.attrs.passcode === '')
                {
                    //variable is undefined, null, or unassigned so unlock the control component
                    component.attrs.unlocked = true;
                    //console.log(selectedId + ' is unlocked');
                }
                //if a passcode is defined for this component and the component is locked...
                else if (component.attrs.unlocked === false && initLatchedSwitches === false)
                {   
                    //A valid passcode has not been entered or has timed out.
                    //Request a passcode for this control.
                    ShowPasscodeDialogBox(component.attrs.id);
                    //console.log(selectedId + ' is locked');
                }
            }
            
            function ShowComponentPosition(component)
            {
                document.getElementById("position_display_textbox").value = "  x = " + component.getPosition().x + ", y = " + component.getPosition().y;
            }
            
            function ClearPositionDisplay()
            {
                document.getElementById("position_display_textbox").value = "";
                //console.log("Cleared position display");
            }
             
            function ComponentProperties(componentObj)
            {                    
                //Build an html string for the component properties dialog box.
                //This string will set the innerHTML of the properties dialog box.             
                if (designMode && selectedId !== "")
                {
                    var propertiesStr = "";
                    var trackStr = "";
                    var toggleSw = false;                 

                    //create common properties options
                    propertiesStr = '<p style="margin-top: 10px">'
                        + 'X Position:&nbsp&nbsp<input id="x_pos_value" type="number" style="width: 208px; float: right" onchange="UpdateComponentPosition(\'' + componentObj.attrs.id + '\')"/></p>'
                        + '<p style="margin-top: 10px">'
                        + 'Y Position:&nbsp&nbsp<input id="y_pos_value" type="number" style="width: 208px; float: right" onchange="UpdateComponentPosition(\'' + componentObj.attrs.id + '\')"/></p>'
                        + '<p style="margin-top: 10px">'     
                        + 'Layer:&nbsp&nbsp<input id="z_value_front" type="button" value="Front" style="width: 53px; float: right" onclick="BringComponentToFront(\'' + componentObj.attrs.id + '\')"/>'
                        + '<input id="z_value_forward" type="button" value="Fwd" style="width: 53px; float: right" onclick="BringComponentForward(\'' + componentObj.attrs.id + '\')"/>' 
                        + '<input id="z_value_backward" type="button" value="Bkwd" style="width: 53px; float: right" onclick="SendComponentBackward(\'' + componentObj.attrs.id + '\')"/>'
                        + '<input id="z_value_back" type="button" value="Back" style="width: 53px; float: right" onclick="SendComponentToBack(\'' + componentObj.attrs.id + '\')"/></p>';                  

                    //if the component is a quadrilateral, create point coordinate property options
                    if (componentObj.attrs.type === "quadrilateral")
                    {
                        //create properties options for quadrilateral shape
                        propertiesStr += '<p style="margin-top: 10px">'
                            + 'Point 1 (x,y):&nbsp&nbsp<input id="point1y_value" type="number" style="width: 103px; float: right" onchange="UpdateQuadrilateralCoordinates(\'' + componentObj.attrs.id + '\')"/>'
                            + '<input id="point1x_value" type="number" style="width: 103px; float: right" onchange="UpdateQuadrilateralCoordinates(\'' + componentObj.attrs.id + '\')"/></p>'                                           
                            + '<p style="margin-top: 10px">'
                            + 'Point 2 (x,y):&nbsp&nbsp<input id="point2y_value" type="number" style="width: 103px; float: right" onchange="UpdateQuadrilateralCoordinates(\'' + componentObj.attrs.id + '\')"/>'
                            + '<input id="point2x_value" type="number" style="width: 103px; float: right" onchange="UpdateQuadrilateralCoordinates(\'' + componentObj.attrs.id + '\')"/></p>'
                            + '<p style="margin-top: 10px">'
                            + 'Point 3 (x,y):&nbsp&nbsp<input id="point3y_value" type="number" style="width: 103px; float: right" onchange="UpdateQuadrilateralCoordinates(\'' + componentObj.attrs.id + '\')"/>'
                            + '<input id="point3x_value" type="number" style="width: 103px; float: right" onchange="UpdateQuadrilateralCoordinates(\'' + componentObj.attrs.id + '\')"/></p>'
                            + '<p style="margin-top: 10px">'
                            + 'Point 4 (x,y):&nbsp&nbsp<input id="point4y_value" type="number" style="width: 103px; float: right" onchange="UpdateQuadrilateralCoordinates(\'' + componentObj.attrs.id + '\')"/>'
                            + '<input id="point4x_value" type="number" style="width: 103px; float: right" onchange="UpdateQuadrilateralCoordinates(\'' + componentObj.attrs.id + '\')"/></p>';
                    }

                    //for a quadrilateral or arrow component, add the rotation property                      
                    if (componentObj.attrs.type === "quadrilateral"
                        || componentObj.attrs.type === "arrow")
                    {
                        propertiesStr += '<p style="margin-top: 10px">'                 
                            + 'Rotation:&nbsp&nbsp<input id="rotation_value" type="number" style="width: 208px; float: right" onchange="RotateComponent(\'' + componentObj.attrs.id + '\')"/></p>';   
                    }

                    if (componentObj.attrs.type !== "text"
                        && componentObj.attrs.type !== "timer")
                    {                                             
                        propertiesStr += '<p style="margin-top: 10px">'
                            + 'Scale:&nbsp&nbsp<input id="scale_value" type="number" style="width: 208px; float: right" onchange="UpdateScaleValue(\'' + componentObj.attrs.id + '\')"/></p>';
                    }

                    //if the component is not text, add line properties
                    if (componentObj.attrs.type === "LED"
                        || componentObj.attrs.type === "quadrilateral"
                        || componentObj.attrs.type === "arrow")
                    {
                        propertiesStr += '<p style="margin-top: 10px">'
                            + 'Line Width:&nbsp&nbsp<input id="line_width_value" type="number" style="width: 208px; float: right" onchange="UpdateLineWidth(\'' + componentObj.attrs.id + '\')"/></p>'
                            + '<p style="margin-top: 10px">'
                            + 'Line Color:&nbsp&nbsp<td><select id="line_color_value" style="width: 210px; float: right" onchange="UpdateLineColor(\'' + componentObj.attrs.id + '\')"></p>';

                        //create line color options
                        for (var i=0; i < iconColors.length; i++)
                        {
                            propertiesStr += '<option>' + iconColors[i] + '</option>';
                        }

                        propertiesStr += '</select></td></div>'
                            + '<p style="margin-top: 10px">'
                            + 'Default Fill Color:&nbsp&nbsp<td><select id="default_fill_color" style="width: 210px; float: right" onchange="UpdateFillColor(\'' + componentObj.attrs.id + '\')"></p>';

                        //create fill color options
                        for (var i=0; i < iconColors.length; i++)
                        {
                            propertiesStr += '<option>' + iconColors[i] + '</option>';
                        }

                        propertiesStr += '</select></td></div>';
                    }

                    if (componentObj.attrs.type === "text" || componentObj.attrs.type === "pb_square")
                    {
                        propertiesStr += '<p style="margin-top: 10px">'
                            + 'Text:&nbsp&nbsp<input id="text_value" type="text" style="width: 208px; float: right" onchange="ChangeTextValue(\'' + componentObj.attrs.id + '\')"/></p>'; 
                    }

                    if (componentObj.attrs.type === "text"
                        || componentObj.attrs.type === "timer"
                        || componentObj.attrs.type === "pb_square")
                    {
                        propertiesStr += '<p style="margin-top: 10px">'
                            + 'Text Size:&nbsp<td><select id="text_size" style="width: 210px; float: right" onchange="ChangeTextSize(\'' + componentObj.attrs.id + '\')"></p>';

                        //create text size options
                        for (var i=10; i <= 44; i=i+2)
                        {
                            propertiesStr += '<option>' + i + '</option>';
                        }

                        if (componentObj.attrs.type === "text"
                            || componentObj.attrs.type === "timer")
                        {
                            propertiesStr += '</select></td></div>'
                                + '<p style="margin-top: 10px">'                 
                                + 'Rotation:&nbsp&nbsp<input id="rotation_value" type="number" style="width: 208px; float: right" onchange="RotateComponent(\'' + componentObj.attrs.id + '\')"/></p>'
                                + '<p style="margin-top: 10px">'
                                + 'Default Text Color:&nbsp&nbsp<td><select id="text_color" style="width: 210px; float: right" onchange="ChangeDefaultTextColor(\'' + componentObj.attrs.id + '\')"></p>';
                        }

                        if (componentObj.attrs.type === "pb_square")
                        {
                            propertiesStr += '</select></td></div>'
                                + '<p style="margin-top: 10px">'                 
                                + 'Text Color:&nbsp&nbsp<td><select id="text_color" style="width: 210px; float: right" onchange="ChangeTextColor(\'' + componentObj.attrs.id + '\')"></p>';
                        }

                        //create text color options
                        for (var i=0; i < iconColors.length; i++)
                        {
                            propertiesStr += '<option>' + iconColors[i] + '</option>';
                        }

                        propertiesStr += '</select></td></div>';

                    }

                    if (componentObj.attrs.type === "timer")
                    {
                        propertiesStr += '<p style="margin-top: 10px">'                 
                            + 'Timer Duration:&nbsp&nbsp<input id="timer_duration_value" type="number" style="width: 208px; float: right" onchange="UpdateTimerDuration(\'' + componentObj.attrs.id + '\')"/></p>';   
                    }

                    //If this is a toggle switch, show option for setting latching or momentary     
                    if (componentObj.attrs.type === "togSw"
                        || componentObj.attrs.type === "togSwMom")
                    {
                        var momentary = "";
                        var rotated = "";

                        if (componentObj.attrs.type === "togSwMom")
                        {
                           momentary = "checked"; 
                        }

                        if (componentObj.attrs.rotation == 90)
                        {
                            rotated = "checked";
                        }

                        propertiesStr += '<p style="margin-top: 10px">'
                            + 'Momentary (ON-OFF-ON):&nbsp&nbsp<input type="checkbox"' + momentary + ' style="width: 20px; float: right" onchange="SetToggleSwitchType(\'' + componentObj.attrs.id + '\',this.checked)"/></p>'
                            + '<p style="margin-top: 10px">'
                            + 'Rotate 90 degrees:&nbsp&nbsp<input type="checkbox"' + rotated + ' style="width: 20px; float: right" onchange="RotateToggleSwitch(\'' + componentObj.attrs.id + '\',this.checked)"/></p>';

                        toggleSw = true;
                    }

                    //if has controls, add passcode property
                    if (componentObj.attrs.nbrCtrls > 0)
                    {
                        propertiesStr += '<p style="margin-top: 10px">'
                            + 'Passcode:&nbsp&nbsp<input id="passcode_value" type="text" style="width: 208px; float: right" onchange="SetPasscode(\'' + componentObj.attrs.id + '\')"/></p>'   
                            + '<p style="margin-top: 10px">'                 
                            + 'Passcode Timeout (s):&nbsp&nbsp<input id="passcode_timeout_value" type="number" style="width: 208px; float: right" onchange="SetPasscodeTimeout(\'' + componentObj.attrs.id + '\')"/></p>';         
                    }

                    //Loop through the number of controls for this component to create its control properties.
                    for (var i=0; i < componentObj.attrs.nbrCtrls ; i++)
                    {
                        var labelNbr = i + 1;
                        var descStr = "";

                        if (toggleSw)
                        {
                            if (labelNbr == 1)
                            {
                                descStr = " (Up/Right Position)";
                            }
                            if (labelNbr == 2)
                            {
                                descStr = " (Down/Left Position)";
                            }
                        }

                        if (componentObj.attrs.type === "keyswitch")
                        {
                            if (labelNbr == 1)
                            {
                                descStr = " (Up Position)";
                            }
                            if (labelNbr == 2)
                            {
                                descStr = " (Right Position)";
                            }
                            if (labelNbr == 3)
                            {
                                descStr = " (Left Position)";
                            }
                        }

                        //the last control is used for the unlock status
                        if (labelNbr == componentObj.attrs.nbrCtrls)
                        {
                            descStr = " (Unlock Status)";
                        }

                        propertiesStr += '<p style="margin-top: 10px">Control Status ' + labelNbr + descStr + ':&nbsp&nbsp'
                            + '<td><select id="control' + i + '_label_options" onchange="SetComponentControlProperty(\'' + componentObj.attrs.id + '\',\'control' + i + '\')" style="float: right; width: 210px">' 
                            + controlOptionsStr
                            + '</select></td></p>';
                    }    

                    if (componentObj.attrs.hasIndications)
                    {
                        //Add flash rate property
                        propertiesStr += '<p style="margin-top: 10px">'
                        + 'Flash Time (ms):&nbsp&nbsp<input id="flash_time_value" type="number" style="width: 208px; float: right" onchange="UpdateComponentFlashTime(\'' + componentObj.attrs.id + '\')"/></p>'

                        //Add the labels and drop down lists to select status names for the colored indication. Also add the lists to select the state option.
                        //The state option determines whether the color is displayed when the bit state is true or when it is false.  
                        var indNbr = 0;
                        for (var i=0; i < iconColors.length; i++)
                        {
                            propertiesStr += '<p style="margin-top: 10px">' 
                                + trackStr + 'Indication (' + iconColors[i] + '):&nbsp&nbsp'
                                + '<td><select id="indication'+ indNbr + '_state_options" onchange="SetComponentIndicationProperty(\'' + componentObj.attrs.id + '\',\'indication' + indNbr + '\')" style="float: right; width: 60px">' 
                                + '<option>True</option>'
                                + '<option>False</option>'
                                + '</select></td>'
                                + '<td><select id="indication'+ indNbr + '_label_options" onchange="SetComponentIndicationProperty(\'' + componentObj.attrs.id + '\',\'indication' + indNbr + '\')" style="float: right; width: 150px">' 
                                + indicationOptionsStr
                                + '</select></td></p>';

                             indNbr += 1;
                        }

                        //Now add the options for flash bits
                         for (var i=0; i < iconColors.length; i++)
                        {
                            propertiesStr += '<p style="margin-top: 10px">' 
                                + trackStr + 'Indication Flashing (' + iconColors[i] + '):&nbsp&nbsp'
                                + '<td><select id="indication'+ indNbr + '_state_options" onchange="SetComponentIndicationProperty(\'' + componentObj.attrs.id + '\',\'indication' + indNbr + '\')" style="float: right; width: 60px">' 
                                + '<option>True</option>'
                                + '<option>False</option>'
                                + '</select></td>'
                                + '<td><select id="indication'+ indNbr + '_label_options" onchange="SetComponentIndicationProperty(\'' + componentObj.attrs.id + '\',\'indication' + indNbr + '\')" style="float: right; width: 150px">' 
                                + indicationOptionsStr
                                + '</select></td></p>';

                             indNbr += 1;
                        }

                    }

                    //Build the final string and assign it to the properties dialog box
                    document.getElementById("properties_dialog").innerHTML = propertiesStr;

                    //Now that we have created the inner html for the properties,
                    //get the current property values for this component.

                    document.getElementById("x_pos_value").value = componentObj.attrs.x;
                    document.getElementById("y_pos_value").value = componentObj.attrs.y;

                    if (componentObj.attrs.type === "text" || componentObj.attrs.type === "pb_square")
                    {
                        document.getElementById("text_value").value = componentObj.find('Text')[0].attrs.text;
                    }

                    if (componentObj.attrs.type === "text" || componentObj.attrs.type === "timer")
                    {
                        document.getElementById("text_color").value = componentObj.find('Text')[0].attrs.defaultColor;
                        document.getElementById("text_size").value = componentObj.find('Text')[0].attrs.fontSize;
                    }
                    else
                    {
                        document.getElementById("scale_value").value = componentObj.attrs.scaleX * 100;   
                    }

                    if (componentObj.attrs.type === "pb_square")
                    {
                        document.getElementById("text_color").value = componentObj.find('Text')[0].attrs.fill;
                        document.getElementById("text_size").value = componentObj.find('Text')[0].attrs.fontSize;
                    }

                    if (componentObj.attrs.type === "quadrilateral")
                    {
                        var quad = componentObj.find('.indicator')[0].attrs;
                        document.getElementById("point1x_value").value = quad.point1x;  
                        document.getElementById("point1y_value").value = quad.point1y;  
                        document.getElementById("point2x_value").value = quad.point2x;  
                        document.getElementById("point2y_value").value = quad.point2y;  
                        document.getElementById("point3x_value").value = quad.point3x;  
                        document.getElementById("point3y_value").value = quad.point3y;  
                        document.getElementById("point4x_value").value = quad.point4x;  
                        document.getElementById("point4y_value").value = quad.point4y;  
                    }

                    if (componentObj.attrs.type === "timer")
                    {
                        document.getElementById("timer_duration_value").value = componentObj.find('Text')[0].attrs.timerDuration;
                    }

                    if (componentObj.attrs.type === "text"
                        || componentObj.attrs.type === "quadrilateral"
                        || componentObj.attrs.type === "arrow"
                        || componentObj.attrs.type === "timer")
                    {
                        document.getElementById("rotation_value").value = componentObj.attrs.rotation;
                    }

                    if (componentObj.attrs.type === "LED"
                        || componentObj.attrs.type === "quadrilateral"
                        || componentObj.attrs.type === "arrow")
                    {
                        document.getElementById("line_width_value").value = componentObj.find('.indicator')[0].attrs.strokeWidth;   
                        document.getElementById("line_color_value").value = componentObj.find('.indicator')[0].attrs.stroke;   
                        document.getElementById("default_fill_color").value = componentObj.find('.indicator')[0].attrs.defaultColor;
                    }

                    if (componentObj.attrs.nbrCtrls >0)
                    {
                        document.getElementById("passcode_value").value = componentObj.attrs.passcode;   
                        document.getElementById("passcode_timeout_value").value = componentObj.attrs.timeout;   
                    }

                    for (var i=0; i < componentObj.attrs.nbrCtrls; i++)
                    {
                        document.getElementById("control" + i + "_label_options").selectedIndex = componentObj.attrs["control" + i];
                    }

                    if (componentObj.attrs.hasIndications)
                    {
                        document.getElementById("flash_time_value").value = componentObj.attrs.flashTime;

                        for (var i=0; i < iconColors.length * 2; i++)
                        {
                            document.getElementById("indication" + i + "_label_options").selectedIndex = componentObj.attrs["indication" + i].split(",")[0];
                            document.getElementById("indication" + i + "_state_options").selectedIndex = componentObj.attrs["indication" + i].split(",")[1];
                        } 
                    }                             
                }
            }
            
            function ShowPropertiesDialog()
            {
                $( "#properties_dialog" ).dialog( "option", "title", "Properties" );
                      
                //Open the properties dialog box                     
                $( "#properties_dialog" ).dialog( "open" );  
                
                    //Request the control labels from the QLCP GUI Developer tool and create the control status name options.
                    $.get("http://localhost:9999/control_labels/", function(data){
                    
                        UpdateControlLabels(data);
                    
                        //Request the indication labels from the QLCP GUI Developer tool and create the indication status name options.
                        $.get("http://localhost:9999/indication_labels/", function(data){
                        
                            UpdateIndicationLabels(data);
                        
                        }, "text");
                    }, "text");   
                
                //console.log("Selected ID = " + selectedId);
                
                if (selectedId !== "")
                {
                    ComponentProperties(stage.get('#'+ selectedId)[0]);
                }
            }
            
            function ClearPreviousSelection()
            {
                //Deselect the previous component
                if (selectedId !== "")
                {   
                    prevObj = stage.get('#'+ selectedId)[0];
                    prevObj.attrs.selected = false;
                    prevObj.find('.selection')[0].attrs.hide();
                    component_layer.draw();
                }
                selectedId = "";
                
                ClearPositionDisplay();
                
                //clear properties dialog box
                document.getElementById("properties_dialog").innerHTML = "";               
            }
            
            //this function shows the selected component and clears the selection
            //of the previously selected component. It also adds or removes a selected
            //component from a group selection. This function depends on requestAnimationFrame
            //to call UpdateComponentLayer in order to see the changes.
            function ShowSelectedComponent(componentObj)
            {
                ClearPreviousSelection();

                if (typeof componentObj !== 'undefined')
                {
                    selectedId = componentObj.attrs.id;
                    ShowComponentPosition(componentObj);
                    
                    //show the component as selected
                    componentObj.find('.selection')[0].attrs.show();

                    if ($( "#properties_dialog" ).dialog( "isOpen" ))
                    {
                        ComponentProperties(componentObj);
                    }
                                               
                    //If there is a group of selected components...
                    if (selectedComponents.length > 0)
                    {
                        //If the Ctrl key is down...
                        if (ctrlKeyDown)
                        {
                            //add the selected component to the selected group if not already selected
                            if (!componentObj.attrs.selected)
                            {
                                console.log("add to selection");
                                componentObj.attrs.selected = true;
                                selectedComponents.push(componentObj);
                            }
                            else //remove the component from the group
                            {
                                var index = selectedComponents.indexOf(componentObj);
                                console.log("remove index " + index + " from selection");
                                if (index > -1)
                                {
                                  selectedComponents.splice(index, 1);
                                }
                                componentObj.attrs.selected = false;
                                ClearPreviousSelection();
                            }
                            compInSelectedGroup = true;
                        }
                    } 
                } 
                   
                console.log("Selected id = " + selectedId);
            }
            
            function SetToggleSwitchType(id, checked)
            {
                var componentObj = stage.get('#'+id)[0];
                
                if (checked)
                {
                    componentObj.attrs.type = "togSwMom";
                    componentObj.find('.up')[0].visible(false);
                    componentObj.find('.down')[0].visible(false);
                    componentObj.find('.center')[0].visible(true);
                }
                else
                {
                    componentObj.attrs.type = "togSw";
                    componentObj.find('.up')[0].visible(false);
                    componentObj.find('.down')[0].visible(true);
                    componentObj.find('.center')[0].visible(false);
                }
                
                component_layer.draw();
            }
            
            function SetComponentControlProperty(id, property)
            {
                //This function is called when control status name (bit assignement) is changed to update the saved values.
                var statusListId = property + "_label_options";
                var componentObj = stage.get('#'+id)[0];
                componentObj.attrs[property] = document.getElementById(statusListId).selectedIndex;
            }
            
            function SetComponentIndicationProperty(id, property)
            {
                //This function is called when an indication status name or state option is changed to update the saved values.
                //The state option determines whether the color is displayed when the bit state is true or when it is false.
                var statusListId = property + "_label_options";
                var stateListId = property + "_state_options";
                var textObj = stage.get('#'+id)[0];
                textObj.attrs[property] = document.getElementById(statusListId).selectedIndex + "," + document.getElementById(stateListId).selectedIndex;
            }
            
            function ChangeTextValue(id)
            {
                var textObj = stage.get('#'+id)[0].find('Text')[0];
                textObj.setText(document.getElementById("text_value").value);
                
                if (textObj.getParent().attrs.type === 'pb_square')
                {
                    textObj.attrs.alignText(); 
                }
                                                                                                         
                //update the selection border to account for line width
                stage.get('#'+id)[0].find('.selection')[0].attrs.update();
                    
                component_layer.draw();
            }
          
            function ChangeDefaultTextColor(id)
            {
                var textObj = stage.get('#'+id)[0].find('Text')[0];
                textObj.attrs.defaultColor = document.getElementById("text_color").value;
                component_layer.draw();
            }
             
            function ChangeTextColor(id)
            {
                var textObj = stage.get('#'+id)[0].find('Text')[0];
                textObj.attrs.fill = document.getElementById("text_color").value;
                component_layer.draw();
            }
            
            function ChangeTextSize(id)
            {
                var textObj = stage.get('#'+id)[0].find('Text')[0];
                textObj.attrs.fontSize = document.getElementById("text_size").value;
                
                //refresh the text value so that selection border resizing will work
                textObj.setText(textObj.attrs.text);

                if (textObj.getParent().attrs.type === 'pb_square')
                {
                    textObj.attrs.alignText(); 
                }
                                                                                                       
                //update the selection border to account for line width
                stage.get('#'+id)[0].find('.selection')[0].attrs.update();
                    
                component_layer.draw();
            }
            
            function IncreaseBackgroundImageSize()
            {
                //if there is an image, resize it based on the current scale factor
                if (document.getElementById('lcp_background').childElementCount > 0)
                {
                    document.getElementById('lcp_background').firstChild.width = Number(document.getElementById('lcp_background').firstChild.width) + 10;
                }
            }
                                    
            function DecreaseBackgroundImageSize()
            {
                //if there is an image, resize it based on the current scale factor
                if (document.getElementById('lcp_background').childElementCount > 0)
                {
                    document.getElementById('lcp_background').firstChild.width = Number(document.getElementById('lcp_background').firstChild.width) - 10;
                }
            }
            
            function RemoveBackgroundImage()
            {
                if (document.getElementById('lcp_background').childElementCount > 0)
                {
                    document.getElementById('lcp_background').removeChild(document.getElementById('lcp_background').firstChild);
                }
            }
                            
            function UpdateComponentPosition(id)
            {
                var component = stage.get('#'+id)[0];

                component.position({
                    x: Math.floor(Number(document.getElementById("x_pos_value").value)),
                    y: Math.floor(Number(document.getElementById("y_pos_value").value))
                });  
                component_layer.draw();
                //console.log("x = " + component.attrs.x + ", y= " + component.attrs.y);
                ShowComponentPosition(component);
            }
            
            function BringComponentForward(id)
            {
                var component = stage.get('#'+id)[0];
                component.setZIndex(component.getZIndex() + 1);
                component_layer.draw();
            }
            
            function SendComponentBackward(id)
            {
                var component = stage.get('#'+id)[0];
                var z = component.getZIndex() - 1;
                if (z < 0)
                {
                    z = 0;
                }
                component.setZIndex(z);
                component_layer.draw();
            }

            function BringComponentToFront(id)
            {
                //find the maximum z index.
                var maxZ = 0;
                for (var j=0; j < component_layer.children.length; j++)
                {                    
                    var cmpObj = component_layer.children[j]; 
                    if (cmpObj.getZIndex() > maxZ)
                    {
                        maxZ = cmpObj.getZIndex();
                    } 
                    console.log(maxZ);
                }
                
                var component = stage.get('#'+id)[0];
                component.setZIndex(maxZ);
                component_layer.draw();
            }
            
            function SendComponentToBack(id)
            {
                var component = stage.get('#'+id)[0];
                component.setZIndex(0);
                component_layer.draw();
            }
            
            function UpdateLineWidth(id)
            {
                if (Number.isInteger(Number(document.getElementById("line_width_value").value)))
                {
                    var component = stage.get('#'+id)[0].find('.indicator')[0];

                    if (document.getElementById("line_width_value").value <= 0)
                    {
                        document.getElementById("line_width_value").value = 0;        
                        component.strokeEnabled(false);
                    }
                    else
                    {
                        component.strokeEnabled(true);
                    }

                    if (document.getElementById("line_width_value").value > 20)
                    {
                        document.getElementById("line_width_value").value = 20;        
                    }

                    component.strokeWidth(document.getElementById("line_width_value").value);                
                                                                                               
                    //update the selection border to account for line width
                    stage.get('#'+id)[0].find('.selection')[0].attrs.update();
                    
                    component_layer.draw();                  
                }
            }           
            
            function UpdateLineColor(id)
            {
                var component = stage.get('#'+id)[0].find('.indicator')[0];
                component.stroke(document.getElementById("line_color_value").value);
                component_layer.draw();
            }           
            
            function UpdateFillColor(id)
            {
                var component = stage.get('#'+id)[0].find('.indicator')[0];
                component.attrs.defaultColor = document.getElementById("default_fill_color").value;
                component_layer.draw();
            }    
            
            function UpdateScaleValue(id)
            {
                var component = stage.get('#'+id)[0];
                 
                component.scale({
                  x: document.getElementById("scale_value").value/100,
                  y: document.getElementById("scale_value").value/100
                });
                
                //update the selection border to account for line width
                stage.get('#'+id)[0].find('.selection')[0].attrs.update();
                    
                component_layer.draw();
            }  
            
            function UpdateQuadrilateralCoordinates(id)
            {
                var minPixelDiff = 2;
                
                //get the quad shape from the stage
                var quad = stage.get('#'+id)[0].find('.indicator')[0].attrs;

                //check coordinates...
                //first, round any decimal values
                document.getElementById("point1x_value").value = Math.round(document.getElementById("point1x_value").value);      
                document.getElementById("point2x_value").value = Math.round(document.getElementById("point2x_value").value);      
                document.getElementById("point3x_value").value = Math.round(document.getElementById("point3x_value").value);      
                document.getElementById("point4x_value").value = Math.round(document.getElementById("point4x_value").value);      
                document.getElementById("point1y_value").value = Math.round(document.getElementById("point1y_value").value);      
                document.getElementById("point2y_value").value = Math.round(document.getElementById("point2y_value").value);      
                document.getElementById("point3y_value").value = Math.round(document.getElementById("point3y_value").value);      
                document.getElementById("point4y_value").value = Math.round(document.getElementById("point4y_value").value);                   
 
                //2x cannot be less than minPixelDiff greater than 1x
                if (Number(document.getElementById("point2x_value").value) - Number(document.getElementById("point1x_value").value) < minPixelDiff)
                {
                    document.getElementById("point2x_value").value = Number(document.getElementById("point1x_value").value) + minPixelDiff;
                }
                                
                //3x cannot be less than minPixelDiff greater than 4x
                if (Number(document.getElementById("point3x_value").value) - Number(document.getElementById("point4x_value").value) < minPixelDiff)
                {
                    document.getElementById("point3x_value").value = Number(document.getElementById("point4x_value").value) + minPixelDiff;
                }
                                                                
                //4y cannot be less than minPixelDiff greater than 1y
                if (Number(document.getElementById("point4y_value").value) - Number(document.getElementById("point1y_value").value) < minPixelDiff)
                {
                    document.getElementById("point4y_value").value = Number(document.getElementById("point1y_value").value) + minPixelDiff;
                }
                                                                
                //3y cannot be less than minPixelDiff greater than 2y
                if (Number(document.getElementById("point3y_value").value) - Number(document.getElementById("point2y_value").value) < minPixelDiff)
                {
                    document.getElementById("point3y_value").value = Number(document.getElementById("point2y_value").value) + minPixelDiff;
                }
                
                quad.point1x = Number(document.getElementById("point1x_value").value);
                quad.point1y = Number(document.getElementById("point1y_value").value);
                quad.point2x = Number(document.getElementById("point2x_value").value);                
                quad.point2y = Number(document.getElementById("point2y_value").value);
                quad.point3x = Number(document.getElementById("point3x_value").value);
                quad.point3y = Number(document.getElementById("point3y_value").value);
                quad.point4x = Number(document.getElementById("point4x_value").value);
                quad.point4y = Number(document.getElementById("point4y_value").value);
                
                //update the selection border since the size has changed
                stage.get('#'+id)[0].find('.selection')[0].attrs.update();
                
                component_layer.draw();
            }
            
            function RotateComponent(id)
            {
                var component = stage.get('#'+id)[0];
                               
                if ((document.getElementById("rotation_value").value < 0)||(document.getElementById("rotation_value").value >= 360))
                {
                    document.getElementById("rotation_value").value = 0;
                }
                
                component.rotation(document.getElementById("rotation_value").value);
                component_layer.draw();
            }
                                    
            function RotateToggleSwitch(id, checked)
            {
                var component = stage.get('#'+id)[0];  
                
                if (checked)
                {
                    component.rotation(90);
                }
                else
                {
                    component.rotation(0);  
                }
                
                component_layer.draw();
            }
            
            function UpdateComponentFlashTime(id)
            {
                var component = stage.get('#'+id)[0];
                
                if (document.getElementById("flash_time_value").value < 0)
                {
                    document.getElementById("flash_time_value").value = 0;
                }
                
                if (document.getElementById("flash_time_value").value > 65535)
                {
                    document.getElementById("flash_time_value").value = 65535;
                }
                
                component.attrs.flashTime = document.getElementById("flash_time_value").value;
            }
                        
            function UpdateTimerDuration(id)
            {
                var component = stage.get('#'+id)[0].find('.indicator')[0];
                
                if (document.getElementById("timer_duration_value").value < 0)
                {
                    document.getElementById("timer_duration_value").value = 0;
                }
                
                if (document.getElementById("timer_duration_value").value > 65535)
                {
                    document.getElementById("timer_duration_value").value = 65535;
                }
                
                component.attrs.timerDuration = document.getElementById("timer_duration_value").value;
                component.setText(component.attrs.timerDuration.toString() + 's');
                                                                                                                                       
                //update the selection border to account for line width
                stage.get('#'+id)[0].find('.selection')[0].attrs.update();
            }
            
            function Duplicate()
            {
                var duplicateComponents = new Array();
                
                DeselectDropComponentButtons();
                
                //Check for single or multiple-component selection
                if (selectedComponents.length === 0)
                {
                    var cmp = stage.get('#'+ selectedId)[0];               
                    DuplicateComponent(cmp, false);
                }
                else //multiple components are selected
                {
                    for (i = 0; i < selectedComponents.length; i++)
                    {
                        var duplicate = DuplicateComponent(selectedComponents[i], true);
                        
                        //save duplicate component
                        duplicateComponents.push(duplicate);
                    }
                                                                              
                    //clear original group of selected components
                    ClearSelectedComponents();
                    
                    //Make the duplicate components the currently selected components.
                    selectedComponents = duplicateComponents;
                }
            }
                        
            function DuplicateComponent(component, groupSelected)
            {
                var attributes = new Array();       //This array contains the key/value pairs of attributes of copied component
                var cmpType = component.attrs.type;
                
                //console.log("Duplicate component type " + cmpType);
                
                dropMode[cmpType] = true;
                var duplicate = AddComponent(component.attrs.x, component.attrs.y, ++idNbr);
                              
                //copy all attributes for the component
                for (key in component.attrs)
                {
                    duplicate.attrs[key] = component.attrs[key]; //this makes sure controls and indication bits are copied
                    attributes[key] = component.attrs[key]; //create attributes array to use with SetComponentProperties function.
                    //console.log("Attribute " + key + "=" + component.attrs[key]);
                }

                //copy all the attributes for the component's indicator object
                if (typeof component.find('.indicator')[0] !== 'undefined')
                {
                    for (key in component.find('.indicator')[0].attrs)
                    {
                        attributes[key] = component.find('.indicator')[0].attrs[key];
                    }     
                }
       
                //copy all the attributes for the component's text object
                if (typeof component.find('Text')[0] !== 'undefined')
                {
                    for (key in component.find('Text')[0].attrs)
                    {
                        attributes[key] = component.find('Text')[0].attrs[key];
                    }     
                }
                
                //offset the location of the duplicate component
                duplicate.attrs.x += 50;
                //duplicate.attrs.y += 50;

                //set the new id
                duplicate.attrs.id = 'c' + idNbr;
                selectedId = duplicate.attrs.id;
                
                SetComponentProperties(duplicate, attributes);
                //console.log("Duplicated component " + component.attrs.id + "; New component id = " +
                //            duplicate.attrs.id + "; Selected id = " + selectedId);

                if (groupSelected === false)
                {
                    ShowSelectedComponent(duplicate);
                    ComponentProperties(duplicate);     
                }
                else
                {
                    duplicate.find('.selection')[0].attrs.show();
                }

                //update the selection border
                if (typeof duplicate.find('Text')[0] !== 'undefined') duplicate.find('Text')[0].setText(duplicate.find('Text')[0].attrs.text);
                duplicate.find('.selection')[0].attrs.update();
                
                dropMode[cmpType] = false;
                
                return duplicate;
            }
            
            function ShowPasscodeDialogBox(id)
            {                                    
                //Build the html string for the code buttons of the passcode dialog box.
                document.getElementById("passcode_dialog").innerHTML = '<p><input id="btn_1" type="button" value="1" style="font-size: 50px; height: 100px; width: 100px; float: left" onclick="PasscodeDigitEntered(1)"/>'
                + '<input id="btn_2" type="button" value="2" style="font-size: 50px; height: 100px; width: 100px; float: left" onclick="PasscodeDigitEntered(2)"/>'
                + '<input id="btn_3" type="button" value="3" style="font-size: 50px; height: 100px; width: 100px; float: left" onclick="PasscodeDigitEntered(3)"/></p>'
                + '<p><input id="btn_4" type="button" value="4" style="font-size: 50px; height: 100px; width: 100px; float: left" onclick="PasscodeDigitEntered(4)"/>'
                + '<input id="btn_5" type="button" value="5" style="font-size: 50px; height: 100px; width: 100px; float: left" onclick="PasscodeDigitEntered(5)"/>'
                + '<input id="btn_6" type="button" value="6" style="font-size: 50px; height: 100px; width: 100px; float: left" onclick="PasscodeDigitEntered(6)"/></p>'          
                + '<p><input id="btn_7" type="button" value="7" style="font-size: 50px; height: 100px; width: 100px; float: left" onclick="PasscodeDigitEntered(7)"/>'
                + '<input id="btn_8" type="button" value="8" style="font-size: 50px; height: 100px; width: 100px; float: left" onclick="PasscodeDigitEntered(8)"/>'
                + '<input id="btn_9" type="button" value="9" style="font-size: 50px; height: 100px; width: 100px; float: left" onclick="PasscodeDigitEntered(9)"/></p>'
                + '<p><input id="btn_enter" type="button" value="Enter" style="font-size: 20px; height: 100px; width: 100px; float: left" onclick="PasscodeEntered(\'' + id + '\')"/>'
                + '<input id="btn_0" type="button" value="0" style="font-size: 50px; height: 100px; width: 100px; float: left" onclick="PasscodeDigitEntered(0)"/>'
                + '<input id="btn_cancel" type="button" value="Cancel" style="font-size: 20px; height: 100px; width: 100px; float: left" onclick="ClosePasscodeDialog()"/></p>';
        
                $( "#passcode_dialog" ).dialog( "option", "title", "Enter Passcode" );

                //Open the passcode dialog box                     
                $( "#passcode_dialog" ).dialog( "open" ); 
                
                // Remove focus on all buttons within the div with class ui-dialog
                $('.ui-dialog :button').blur();
            }
            
            function PasscodeDigitEntered(digit)
            {
                passcode += digit;
            }
            
            function ClosePasscodeDialog()
            {
                passcode = "";
                $( "#passcode_dialog" ).dialog( "close" );
            }
            
            function PasscodeEntered(id)
            {                
                //check passcode
                var component = stage.get('#'+id)[0];
               
                if (component.attrs.passcode === passcode)
                {
                    component.attrs.unlocked = true;
                    
                    //send the control bit for unlock status
                    var n = component.attrs.nbrCtrls - 1;
                    SendControl(component.attrs["control" + n]-1, 1);
                    
                    $( "#message_dialog" ).dialog( "option", "title", "Passcode Accepted" );

                    document.getElementById("message_dialog").innerHTML = '<div><img style="float: left; margin-top: 5px; width: 36px" src="Images/info32px.png"/>'
                    + '<div style="float: left; margin-left: 15px; margin-top: 6px; width: 350px">This control is now unlocked.<br/><br/></div></div>';   
                                
                    ClosePasscodeDialog(); 
                    
                    //set timeout for passcode
                    setTimeout(PasscodeTimeout, component.attrs.timeout * 1000, id);
                    
                    //console.log(Date() + ' Component ' + id + ' passcode accepted; control unlocked');
                }
                else
                {
                    component.attrs.unlocked = false;
                    
                    $( "#message_dialog" ).dialog( "option", "title", "Incorrect Passcode" );

                    document.getElementById("message_dialog").innerHTML = '<div><img style="float: left; margin-top: 5px; width: 36px" src="Images/warning_icon.png"/>'
                    + '<div style="float: left; margin-left: 15px; margin-top: 6px; width: 350px">This control will remain locked.<br/><br/></div></div>';   
                }
                
                ShowMessageDialog();
               
                passcode = "";
            }
            
            function SetPasscode(id)
            {
                var component = stage.get('#'+id)[0];
                component.attrs.passcode = document.getElementById("passcode_value").value;
            }
            
            function SetPasscodeTimeout(id)
            {
                var component = stage.get('#'+id)[0];
                
                if (document.getElementById("passcode_timeout_value").value < 0)
                {
                    document.getElementById("passcode_timeout_value").value = 0;
                }
                
                if (document.getElementById("passcode_timeout_value").value > 65535)
                {
                    document.getElementById("passcode_timeout_value").value = 65535;
                }
                
                component.attrs.timeout = document.getElementById("passcode_timeout_value").value;
            }
            
            function PasscodeTimeout(id)
            {
                var component = stage.get('#'+id)[0];
                var elapsed = SecondsElapsed(lastClickTime);
                
                //Check time elapsed since last control was sent
                if (elapsed > component.attrs.timeout)
                {
                    //clear the passcode result to force re-entry
                    component.attrs.unlocked = false;
                    
                    //send the control bit for unlock status
                    var n = component.attrs.nbrCtrls - 1;
                    SendControl(component.attrs["control" + n]-1, 0);

                    //console.log(Date() + ' Component ' + id + ' passcode timeout; control locked');
                } 
                else
                {
                    //start the timeout timer again, subtracting the time since last click;
                    //otherwise, the timeout will be this time difference plus the timeout value (too long).
                    var adjustedTimeout = (component.attrs.timeout - elapsed)*1000;
                    setTimeout(PasscodeTimeout, adjustedTimeout, id);
                }            
            }
            
            //This function will delete or move selected component(s) using the keyboard.
            //This function depends on requestAnimationFrame to redraw the component layer.            
            $(document).keydown(function (e){
                                
                if (designMode)
                {
                    //if this is just one selected component...
                    if (selectedId !== "" && selectedComponents.length === 0)
                    {
                        var selObj = stage.get('#'+ selectedId)[0];
                        var positionChanged = false;

                        switch(e.which)
                        {
                            case 46:    //delete

                                ClearPreviousSelection();
                                QnetClientThread.removeEventListener("message", selObj.msgListener, false);
                                selObj.destroy();

                                if ($( "#properties_dialog" ).dialog( "isOpen" ))
                                {
                                    $( "#properties_dialog" ).dialog( "close" );
                                }                                         
                                break;

                            case 37:     //move left if properties dialog box is not open
                                if (!$( "#properties_dialog" ).dialog( "isOpen" ))
                                {
                                    e.preventDefault();
                                    selObj.position({
                                    x:  selObj.attrs.x - 1,
                                    y:  selObj.attrs.y
                                    });
                                    positionChanged = true;
                                }
                                break;

                            case 38:     //move up if properties dialog box is not open
                                if (!$( "#properties_dialog" ).dialog( "isOpen" ))
                                {
                                    e.preventDefault();
                                    selObj.position({
                                        x:  selObj.attrs.x,
                                        y:  selObj.attrs.y - 1
                                        });
                                    positionChanged = true;
                                    break;
                                }

                            case 39:     //move right if properties dialog box is not open
                                if (!$( "#properties_dialog" ).dialog( "isOpen" ))
                                {
                                    e.preventDefault();
                                    selObj.position({
                                        x:  selObj.attrs.x + 1,
                                        y:  selObj.attrs.y
                                        });
                                    positionChanged = true;
                                    break;
                                }

                            case 40:     //move down if properties dialog box is not open
                                if (!$( "#properties_dialog" ).dialog( "isOpen" ))
                                {
                                    e.preventDefault();
                                    selObj.position({
                                        x:  selObj.attrs.x,
                                        y:  selObj.attrs.y + 1
                                        });
                                    positionChanged = true;
                                    break;
                                }

                            case 16:
                                shiftKeyDown = true;
                                break;

                            case 17:
                                ctrlKeyDown = true;
                        }

                        ShowComponentPosition(selObj);

                        if ($( "#properties_dialog" ).dialog( "isOpen" ) && positionChanged)
                        {            
                            document.getElementById("x_pos_value").value = selObj.attrs.x;
                            document.getElementById("y_pos_value").value = selObj.attrs.y;
                        } 
                    }
                    else //there is a selected group of components
                    {
                        switch(e.which)
                        {
                            case 46:    //delete multiple selections (selected group)

                                for (i = 0; i < selectedComponents.length; i++)
                                {
                                    QnetClientThread.removeEventListener("message", selectedComponents[i].msgListener, false);
                                    selectedComponents[i].destroy();
                                }
                                selectedComponents.length = 0;
                                selectedId = "";
                                break;  

                            case 37:     //move group left
                                e.preventDefault();
                                for (i = 0; i < selectedComponents.length; i++)
                                {
                                    selectedComponents[i].position({
                                        x: selectedComponents[i].getPosition().x - 1,
                                        y: selectedComponents[i].getPosition().y
                                    });  
                                }
                                break;

                            case 38:     //move group up
                                e.preventDefault();
                                for (i = 0; i < selectedComponents.length; i++)
                                {
                                    selectedComponents[i].position({
                                        x: selectedComponents[i].getPosition().x,
                                        y: selectedComponents[i].getPosition().y - 1
                                    });  
                                }
                                break;

                            case 39:     //move group right
                                e.preventDefault();
                                for (i = 0; i < selectedComponents.length; i++)
                                {
                                    selectedComponents[i].position({
                                        x: selectedComponents[i].getPosition().x + 1,
                                        y: selectedComponents[i].getPosition().y
                                    });  
                                }
                                break;

                            case 40:     //move group down
                                e.preventDefault();
                                for (i = 0; i < selectedComponents.length; i++)
                                {
                                    selectedComponents[i].position({
                                        x: selectedComponents[i].getPosition().x,
                                        y: selectedComponents[i].getPosition().y + 1
                                    });  
                                }
                                                               
                            case 16:
                                shiftKeyDown = true;
                                break;

                            case 17:
                                ctrlKeyDown = true;
                        }
                    }
                }             
            });
            
            $(document).keyup(function (){
                shiftKeyDown = false;
                ctrlKeyDown = false;
            });

            //When selecting multiple components, this function draws the selection box
            //This function depends on requestAnimationFrame to redraw the component layer.
            function CreateSelectionBox()
            {
                if (draggingToSelectComponents)
                {
                    var mousepos = stage.getPointerPosition();
                    var x = mousepos.x;
                    var y = mousepos.y;

                    if (!selectionBoxCreated)
                    {
                        initX = x;
                        initY = y;

                        //create the selection box
                        selectionBox = new Kinetic.Rect({
                                x: initX,
                                y: initY,
                                height: 1,
                                width: 1,
                                fill: 'transparent',
                                stroke: 'white',
                                strokeWidth: 1
                        });
                        component_layer.add(selectionBox);
                        selectionBoxCreated = true;
                    }
                    else
                    {
                        var height;
                        var width;
                        var newX;
                        var newY;

                        if (x > initX)
                                newX = initX;
                        else
                                newX = x;

                        if (y > initY)
                                newY = initY;
                        else
                                newY = y;

                        height = Math.abs(Math.abs(y) - Math.abs(initY));
                        width = Math.abs(Math.abs(x) - Math.abs(initX));

                        selectionBox.setHeight(height);
                        selectionBox.setWidth(width);
                        selectionBox.setX(newX);
                        selectionBox.setY(newY);                         
                    }
                }
            }
            
            //This function finds the components in the selected area and highlights them.
            function ShowSelectedGroup()
            {    
                selectedComponents.length = 0;

                initX = 0;
                initY = 0;

                var arGroups = component_layer.getChildren();

                for (i = 0; i < arGroups.length; i++)
                {
                    var grp = arGroups[i];

                    if (typeof grp.attrs.id !== 'undefined')
                    {
                        var pos = selectionBox.getAbsolutePosition();

                        //get the extents of the selection box
                        var selRecXStart = parseInt(pos.x);
                        var selRecXEnd = parseInt(pos.x) + parseInt(selectionBox.attrs.width);
                        var selRecYStart = parseInt(pos.y);
                        var selRecYEnd = parseInt(pos.y) + parseInt(selectionBox.attrs.height);

                        //get the extents of the group to compare to
                        var grpXStart = parseInt(grp.attrs.x);
                        var grpXEnd = parseInt(grp.attrs.x) + parseInt(grp.width());
                        var grpYStart = parseInt(grp.attrs.y);
                        var grpYEnd = parseInt(grp.attrs.y) + parseInt(grp.height());
                                        
                        //Are we inside the selction area?
                        if ((selRecXStart <= grpXStart && selRecXEnd >= grpXEnd) && 
                        (selRecYStart <= grpYStart && selRecYEnd >= grpYEnd))
                        {                        
                            grp.find('.selection')[0].attrs.show();
                            grp.attrs.selected = true;
                            selectedComponents.push(grp);
                        }
                    }
                }
            }
                  
            //When multiple components are selected, this function clears the selection highlights.
            //This function depends on requestAnimationFrame to redraw the component layer.
            function ClearSelectedComponents()
            {                
                //for each selected component
                for (i=0; i < selectedComponents.length; i++)
                {                                   
                    selectedComponents[i].find('.selection')[0].attrs.hide();
                    selectedComponents[i].attrs.selected = false;
                }
                
                //clear the array of selected components
                selectedComponents.length = 0;
                redrawComponents = true;
            }
            
            function ToggleFullScreen()
            {

                if ((document.fullscreenElement !== undefined && document.fullscreenElement === null) || 
                    (document.msFullscreenElement !== undefined && document.msFullscreenElement === null) || 
                    (document.mozFullScreenElement !== undefined && !document.mozFullScreenElement) || 
                    (document.webkitFullscreenElement !== undefined && !document.webkitFullscreenElement))
                {
                    //full screen                          
                    if (document.documentElement.requestFullscreen)
                    {
                        document.documentElement.requestFullscreen();
                    }
                    else if (document.documentElement.mozRequestFullScreen)
                    {
                        document.documentElement.mozRequestFullScreen();
                    } 
                    else if (document.documentElement.webkitRequestFullscreen)
                    {
                        document.documentElement.webkitRequestFullscreen(Element.ALLOW_KEYBOARD_INPUT);
                    }
                    else if (document.documentElement.msRequestFullscreen)
                    {
                        document.documentElement.msRequestFullscreen();
                    }                   
                }
                else
                {
                    //exit full screen
                    if (document.exitFullscreen)
                    {
                        document.exitFullscreen();
                    }
                    else if (document.mozCancelFullScreen)
                    {
                        document.mozCancelFullScreen();
                    }
                    else if (document.webkitExitFullscreen)
                    {
                        document.webkitExitFullscreen();
                    }
                    else if (document.msExitFullscreen)
                    {
                        document.msExitFullscreen();
                    }                            
                }          
            }
            
            //Detect a change in full screen mode so that the button image and tool tip can be updated/toggled accordingly.
            //This is required because F11 and the escape key can be used to change the mode, not just the button that calls ToggleFullScreen.
            $(document).on('fullscreenchange webkitfullscreenchange mozfullscreenchange msfullscreenchange', function () {
                
                var fullscreenElement = document.fullscreenElement || document.webkitFullscreenElement || document.mozFullscreenElement || document.msFullscreenElement;

                if (!fullscreenElement) 
                {
                    //toggle button image and tool tip
                    document.getElementById("fullscreen_button").title = "Full Screen";
                    document.getElementById("fullscreen_image").src = "Images/full_screen.png";        
                }
                else
                {
                    //toggle button image and tool tip
                    document.getElementById("fullscreen_button").title = "Exit Full Screen";
                    document.getElementById("fullscreen_image").src = "Images/exit_full_screen.png";                    
                }
            });
 
//************************************************************************************************************************/
                       
        </script>

    </head>

    <body>

        <div id="headerWrapper" onclick="" class="noselect">

            <div id="header">

                <div id="logo">
                    <img src="Images/qlcp_icon.png" style="width: 90%; margin-top: -3px" draggable="false"/>
                </div>

                <div id="title">
                    <h1 id="headerTitle" style="margin-left: 10px">QuEST Local Control Panel</h1>
                </div>

            </div>      
                      
            <img style="float: right; margin-top: 13px; margin-right: 10px" id="link_status" src="Images/error24px.png" draggable="false"/>
            <div style="float: right; margin-top: 18px; margin-right: 5px; margin-left: 30px"><h4>Link</h4></div>
            <div style="float: right; margin-right: 10px"><h1 id="system_time"></h1></div>

        </div>

        <div id="buttonBackground" onclick="" class="noselect">
            
            <div style="position: fixed; top: calc(100vh - 95px); left: calc(100% - 70px); background-color: #222222; padding: 5px" id="fullscreen_button" title="Full Screen" onclick="ToggleFullScreen()"><img id="fullscreen_image" class="buttonImage" src="Images/full_screen.png" align="absmiddle"></div>

            <div id="designButtons" style="float: left; visibility: hidden" onclick="ClearPreviousSelection()">  
                <button class="buttonStyle" id="open_button" title="Open" type="button" onclick="OpenPanelFile()"><img class="buttonImage" src="Images/openfolder24px.png" align="absmiddle"></button>
                <button class="buttonStyle" id="save_button" title="Save" type="button" onclick="SavePanel()"><img class="buttonImage" src="Images/save24px.png" align="absmiddle"></button>
                <button class="buttonStyle" id="clear_all_button" title="Clear All" type="button" onclick="ShowClearAllDialog()"><img class="buttonImage" src="Images/eraser24px.png" align="absmiddle"></button>                
                <button class="buttonStyle" id="run_button" title="Run" type="button" onclick="RunMode()"><img class="buttonImage" src="Images/run24px.png" align="absmiddle"></button>
                <td><h4 style="float: left; margin-left: 30px">Design Style:<select id="design_style_option" onchange="SetPanelStyle(true)" style="width: 140px; margin-top: 13px; margin-left: 5px; margin-right: 8px"><option>Components</option><option>Mosaic Tiles</option></select></h4></td>          
            </div>
            <div id="componentButtons" style="float: left; visibility: hidden">
                <button class="buttonStyle" id="properties_button" title="Show Properties" type="button" onclick="ShowPropertiesDialog()" style="margin-left: 25px"><img class="buttonImage" src="Images/settings.png" align="absmiddle"></button>
                <button class="buttonStyle" id="icons_button" title="Tiles" type="button" onclick="ShowIconsDialog()"><img class="buttonImage" src="Images/icons24px.png" align="absmiddle"></button>
                <button class="buttonStyle" id="text_button" title="Drop Text" type="button" onclick="ToggleDropComponentButtons(id)"><img class="buttonImage" src="Images/text24px.png" align="absmiddle"></button>
                <button class="buttonStyle" id="led_button" title="Drop LED" type="button" onclick="ToggleDropComponentButtons(id)"><img class="buttonImage" src="Images/LED.png" align="absmiddle"></button>
                <button class="buttonStyle" id="pb_button" title="Drop Pushbutton" type="button" onclick="ToggleDropComponentButtons(id)"><img class="buttonImage" src="Images/pushbutton.png" align="absmiddle"></button>
                <button class="buttonStyle" id="pb_square_button" title="Drop Square Pushbutton" type="button" onclick="ToggleDropComponentButtons(id)"><img class="buttonImage" src="Images/square_pb.png" align="absmiddle"></button>
                <button class="buttonStyle" id="togSw_button" title="Drop Toggle Switch" type="button" onclick="ToggleDropComponentButtons(id)"><img class="buttonImage" src="Images/toggle_switch.png" align="absmiddle"></button>
                <button class="buttonStyle" id="keyswitch_button" title="Drop Key Switch" type="button" onclick="ToggleDropComponentButtons(id)"><img class="buttonImage" src="Images/keyswitch.png" align="absmiddle"></button>
                <button class="buttonStyle" id="quadrilateral_button" title="Drop Quadrilateral" type="button" onclick="ToggleDropComponentButtons(id)"><img class="buttonImage" src="Images/Rectangle.png" align="absmiddle"></button>
                <button class="buttonStyle" id="arrow_button" title="Drop Arrow" type="button" onclick="ToggleDropComponentButtons(id)"><img class="buttonImage" src="Images/arrow.png" align="absmiddle"></button>
                <button class="buttonStyle" id="timer_button" title="Drop Timer" type="button" onclick="ToggleDropComponentButtons(id)"><img class="buttonImage" src="Images/Timer_icon.png" align="absmiddle"></button>
                <button class="buttonStyle" id="duplicate_button" title="Duplicate" type="button" onclick="Duplicate()" style="margin-left: 45px"><img class="buttonImage" src="Images/copy.png" align="absmiddle"></button>
            </div>
            <div style="float: left; visibility: hidden">
                <button class="buttonStyle" id="increase_button" title="Increase Background Image Size" type="button" onclick="IncreaseBackgroundImageSize()"><img class="buttonImage" src="Images/image_increase_size.png" align="absmiddle"></button>
                <button class="buttonStyle" id="decrease_button" title="Decrease Background Image Size" type="button" onclick="DecreaseBackgroundImageSize()"><img class="buttonImage" src="Images/image_decrease_size.png" align="absmiddle"></button>
                <button class="buttonStyle" id="remove_background_button" title="Remove Background Image" type="button" onclick="RemoveBackgroundImage()"><img class="buttonImage" src="Images/remove_image.png" align="absmiddle"></button>
                <h4 id="position_display" style="float: right; margin-top: 13px; margin-left: 45px">Position: <input type="text" id="position_display_textbox" readonly style="width: 140px;"></h4>
            </div>

            <div id="runButtons" style="position: absolute; visibility: hidden">  
                <button class="buttonStyle" id="design_button" title="Design" type="button" onclick="DesignMode()"><img class="buttonImage" src="Images/design22px.png" align="absmiddle"></button>
                <button class="buttonStyle" id="send_all_button" title="Send All Controls" type="button" onclick="SendAllControls()"><img class="buttonImage" src="Images/send_controls24px.png" align="absmiddle"></button>
                <button class="buttonStyle" id="clear_controls_button" title="Clear All Controls" type="button" onclick="ClearAllControls()"><img class="buttonImage" src="Images/clear_controls24px.png" align="absmiddle"></button>
            </div>
 
        </div>
                
        <!-- Create the container for scrolling the soft panel within the viewport. -->
        <div id="scroll_container">
            
            <div id="mainContent">

                <!-- Create the container to hold the component layer. -->
                <div id="component_container"></div>

                <div id="lcp_background"></div>

                <!-- Create the divs to hold the grid lines and the icon grid. -->
                <!-- They must be separate to be able to show and hide grid lines and still maintain selection and drag/drop capability -->
                <!-- Note the importance of the z-index attribute in the control-panel.css file. This is what creates the layers for icons, grid lines, and text -->
                <div id="grid_lines"></div>
                <div id="grid"></div>

                <!-- this is a hidden div that is used for building icons during the open panel file process -->
                <div id="temp" style="visibility: hidden"></div> 

                <div id="gridCellSize" class="gridLines" style="visibility: hidden"></div>
                
            </div>
            
        </div>
            
        <!-- create the dialog box divs -->
        <div id="icons_dialog" title="Tiles"></div>
        <div id="properties_dialog" title="Properties"></div>
        <div id="controls_dialog" title="Controls"></div>
        <div id="clear_all_dialog" title="Clear All"></div>
        <div id="message_dialog" title=""></div>
        <div id="passcode_dialog" title="Passcode"></div>

        <div id="footerWrapper" class="noselect">
            <div id="footerContent">
                <h3 style="float: left">&#169; 2019 QuEST Rail LLC., All Rights Reserved</h3>    
                <h4 style="float: right; padding-right: 30px; padding-top: 7px">version 4.8</h4>           
            </div>
        </div>

        <script defer="defer">                                                                                                                                                                                                          
            
            //Create the stage for component and text.
            //46 pixels is based on the icon size, grid size, overlap of grids, etc.

            var stage = new Kinetic.Stage({
                container: 'component_container',
                width: nbrCols * 46,
                height: nbrRows * 46
            });
            
            //create the component layer and add it to the stage
            var component_layer = new Kinetic.Layer();
            stage.add(component_layer);  
            
            var scrollContainer = document.getElementById('scroll_container');
            scrollContainer.addEventListener('scroll', function () {
                var dx = scrollContainer.scrollLeft;
                var dy = scrollContainer.scrollTop;
                stage.container().style.transform = 'translate(' + dx + 'px, ' + dy + 'px)';
                stage.x(-dx);
                stage.y(-dy);
                stage.batchDraw();
            });
        
            //The following three mouse event routines are for selecting multiple components on the stage
            //(on the component layer). The reason the event listeners are bound to the stage is so
            //they only fire when the event occurs in the stage area, not the title, menu, or footer areas.
            
            //This event depends on bubbling of events from the component elements on the component layer.
            //It depends on the component setting the selected Id and mouseDownOnComponent flag first.
            //This routine is only used for design mode.
            stage.getContent().addEventListener("mousedown", function ()
            {
                if (designMode)
                {
                    //If the mouse down event was on the stage and not on a component, and if
                    //there is a drop mode active, add the component to the stage.
                    //AddComponent function will check drop mode to determine which component to add.
                    if (!mouseDownOnComponent)
                    {
                        console.log("clear selected group of components");
                        //Need to clear any previously selected group of components
                        ClearSelectedComponents(); 
                        
                        if (dropMode['none'])
                            draggingToSelectComponents = true;
                        
                        //get the current mouse position
                        var mousepos = stage.getPointerPosition();

                        //If drop mode is not none, add the component to the layer at the current mouse position.
                        if (!dropMode['none'])
                        {
                            var Component = AddComponent(mousepos.x, mousepos.y, ++idNbr);                           
                            ShowSelectedComponent(Component);
                        }
                        else
                        {
                            ClearPreviousSelection();
                        }
                    }
                    else //mouse down was on a component
                    {
                        //If multiple components were selected (array is not empty),
                        //set flag if a component within the group has been selected.
                        for (i = 0; i < selectedComponents.length; i++)
                        {
                            if (selectedComponents[i].attrs.id === selectedId)
                            {
                                compInSelectedGroup = true;
                            }
                        }   
                        //If the mouse down event hit a component in a selected group,
                        //get the mouse position to start keeping track of the move distance (delta x, delta y).
                        if (compInSelectedGroup)
                        {
                            var mousepos = stage.getPointerPosition();
                            initX = mousepos.x;
                            initY = mousepos.y;
                        }
                        else
                        {
                            //Need to clear any previously selected group of components
                            ClearSelectedComponents();  
                        }
                    }                              
                }
            });

            //This function depends on requestAnimationFrame to redraw the component layer.
            //This routine is only used for design mode.
            stage.getContent().addEventListener('mousemove', function ()
            {
                if (designMode)
                {                          
                    if (draggingToSelectComponents)
                    {
                        CreateSelectionBox();
                    }

                    //If a group is selected and a component in the group has been selected,
                    //start moving the group of selected components.
                    if (selectedComponents.length > 0 && compInSelectedGroup)
                    {
                        //Get the current mouse position and determine the delta x and y values.
                        var mousepos = stage.getPointerPosition();
                        deltaX = mousepos.x - initX;
                        deltaY = mousepos.y - initY;

                        //Move each of the components in the selection.
                        for (i = 0; i < selectedComponents.length; i++)
                        {
                            //Move all other components. The selected one will move when dragged.
                            if (selectedComponents[i].attrs.id !== selectedId)
                            {
                                selectedComponents[i].position({
                                    x: selectedComponents[i].getPosition().x + deltaX,
                                    y: selectedComponents[i].getPosition().y + deltaY
                                });  
                            }
                        }
                        
                        //Reset coordinates for the next move event.
                        initX = mousepos.x;
                        initY = mousepos.y;
                    }
                    
                    //show the currently selected component's x,y position
                    if (selectedId !== "")
                    {
                        var component = stage.get('#'+ selectedId)[0];
                        ShowComponentPosition(component);
                        
                        if ($( "#properties_dialog" ).dialog( "isOpen" ))
                        {            
                            document.getElementById("x_pos_value").value = component.attrs.x;
                            document.getElementById("y_pos_value").value = component.attrs.y;
                        } 
                    }
                }
            });

            //This function depends on requestAnimationFrame to redraw the component layer.
            //This routine is only used for design mode.
            stage.getContent().addEventListener("mouseup", function ()
            {
                if (designMode)
                {
                    //When the mouse is released after dragging to select
                    //components (flag is still true), determine which
                    //components are within the selection box and show them as selected.
                    if (draggingToSelectComponents && !mouseDownOnComponent)
                    {
                        draggingToSelectComponents = false;

                        //clear the selection box
                        if (selectionBoxCreated)
                        {
                            ShowSelectedGroup();
                            selectionBox.destroy();
                            selectionBoxCreated = false; 
                        }
                    }
                                       
                    //If there is a group of selected components...
                    if (selectedComponents.length > 0)
                    {
                        //clear the selected Id and previously selected Id.
                        selectedId = "";
                    }  
                                                            
                    //clear flags for next selection
                    compInSelectedGroup = false;
                    mouseDownOnComponent = false; 
                }
            });
            
        </script>        
    </body>
</html>
